# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib.patches as mpatches
from matplotlib.axes import Axes
from matplotlib.image import AxesImage
from matplotlib.projections import register_projection
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib_scalebar import dimension, scalebar
import numpy as np

from orix.scalar import Scalar
from orix.vector import Vector3d


class CrystalMapPlot(Axes):
    """2D plotting of :class:`~orix.crystal_map.crystal_map.CrystalMap`
    objects.
    """

    name = "plot_map"
    _data_axes = None
    _data_slices = None
    _data_shape = None
    colorbar = None
    scalebar = None

    def plot_map(
        self,
        crystal_map,
        value=None,
        scalebar=True,
        scalebar_properties=None,
        legend=True,
        legend_properties=None,
        axes=None,
        depth=None,
        override_status_bar=False,
        **kwargs,
    ) -> AxesImage:
        r"""Plot a 2D map with any CrystalMap attribute as map values.

        Wraps :meth:`matplotlib.axes.Axes.imshow`, see that method for
        relevant keyword arguments.

        Parameters
        ----------
        crystal_map : orix.crystal_map.CrystalMap
            Crystal map object to obtain data to plot from.
        value : numpy.ndarray, optional
            Attribute array to plot. If value is None (default), a phase
            map is plotted.
        scalebar : bool, optional
            Whether to add a scalebar (default is True) along the
            horizontal map dimension.
        scalebar_properties : dict
            Dictionary of keyword arguments passed to
            :func:`mpl_toolkits.axes_grid1.anchored_artists.AnchoredSizeBar`.
        legend : bool, optional
            Whether to add a legend to the plot. This is only
            implemented for a phase plot (in which case default is
            True).
        legend_properties : dict
            Dictionary of keyword arguments passed to
            :meth:`matplotlib.axes.legend`.
        axes : tuple of ints, optional
            Which data axes to plot if data has more than two
            dimensions. The index of data to plot in the final dimension
            is determined by `depth`. If None (default), data along the
            two last axes is plotted.
        depth : int, optional
            Which layer along the third axis to plot if data has more
            than two dimensions. If None (default), data in the first
            index (layer) is plotted.
        override_status_bar : bool, optional
            Whether to display Euler angles and any overlay values in
            the status bar when hovering over the map (default is
            False).
        kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.imshow`.

        Returns
        -------
        im : matplotlib.image.AxesImage
            Image object, to be used further to get data from etc.

        See Also
        --------
        matplotlib.axes.Axes.imshow
        orix.plot.CrystalMapPlot.add_scalebar
        orix.plot.CrystalMapPlot.add_overlay
        orix.plot.CrystalMapPlot.add_colorbar

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> from orix import plot
        >>> from orix.io import load

        Import a crystal map

        >>> xmap = load("/some/directory/data.ang")

        Plot a phase map

        >>> fig = plt.figure()  # Get figure
        >>> ax = fig.add_subplot(projection="plot_map")  # Get axes
        >>> im = ax.plot_map(xmap)  # Get image

        Add an overlay

        >>> ax.add_overlay(xmap, xmap.iq)

        Plot an arbitrary map property, also changing scalebar location

        >>> ax = plt.subplot(projection="plot_map")
        >>> ax.plot_map(
        ...     xmap, xmap.dp, cmap="cividis", scalebar_properties={"loc": 4}
        ... )

        Add a colorbar

        >>> cbar = ax.add_colorbar("Dot product")  # Get colorbar

        Plot orientation angle in degrees of one phase

        >>> xmap2 = xmap["austenite"]
        >>> austenite_angles = xmap2.orientations.angle.data * 180 / np.pi
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(projection="plot_map")
        >>> im = ax.plot_map(xmap2, austenite_angles)
        >>> ax.add_colorbar("Orientation angle [$^{\circ}$]")

        Remove all figure and axes padding

        >>> ax.remove_padding()

        Write annotated figure to file

        >>> fig.savefig(
        ...     "/some/directory/image.png",
        ...     pad_inches=0,
        ...     bbox_inches="tight"
        ... )

        Write un-annotated image to file

        >>> plt.imsave("/some/directory/image2.png", im.get_array())
        """
        self._set_plot_shape(crystal_map=crystal_map, axes=axes, depth=depth)

        patches = None
        if value is None:  # Phase map
            # Color each map pixel with corresponding phase color RGB tuple
            phase_id = crystal_map.get_map_data("phase_id")
            phase_id = phase_id[self._data_slices]
            unique_phase_ids = np.unique(phase_id[~np.isnan(phase_id)])
            data = np.ones(phase_id.shape + (3,))
            for i, color in zip(
                unique_phase_ids, crystal_map.phases_in_data.colors_rgb
            ):
                mask = phase_id == int(i)
                data[mask] = data[mask] * color

            # Add legend patches to plot
            patches = []
            for _, p in crystal_map.phases_in_data:
                patches.append(mpatches.Patch(color=p.color_rgb, label=p.name))
        else:  # Create masked array of correct shape
            if isinstance(value, (Scalar, Vector3d)):
                value = value.data
            data = crystal_map.get_map_data(value)
            data = data[self._data_slices]

        # Squeeze 1-dimensions
        data = np.squeeze(data)

        # Legend
        if legend and isinstance(patches, list):
            if legend_properties is None:
                legend_properties = {}
            self._add_legend(patches, **legend_properties)

        # Scalebar
        if scalebar:
            if scalebar_properties is None:
                scalebar_properties = {}
            _ = self.add_scalebar(crystal_map, **scalebar_properties)

        im = self.imshow(X=data, **kwargs)
        if override_status_bar:
            im = self._override_status_bar(im, crystal_map)

        return im

    def add_scalebar(self, crystal_map, **kwargs):
        """Add a scalebar to the axes instance and return it.

        The scalebar is also available as an attribute :attr:`scalebar`.

        Parameters
        ----------
        crystal_map : orix.crystal_map.CrystalMap
            Crystal map instance to obtain necessary data from.
        kwargs
            Keyword arguments passed to
            :class:`matplotlib_scalebar.scalebar.ScaleBar`.

        Returns
        -------
        bar : matplotlib_scalebar.scalebar.ScaleBar
            Scalebar.

        Examples
        --------
        >>> xmap
        Phase  Orientations   Name       Symmetry  Color
        1      5657 (48.4%)   austenite  432       tab:blue
        2      6043 (51.6%)   ferrite    432       tab:orange
        Properties: iq, dp
        Scan unit: um

        Create a phase map without a scale bar and add it afterwards

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(projection="plot_map")
        >>> im = ax.plot_map(xmap, scalebar=False)
        >>> sbar = ax.add_scalebar(xmap, location=4, frameon=False)
        """
        # Get whether z, y or x
        last_axis = crystal_map.ndim - 1
        horizontal = crystal_map._coordinate_axes[last_axis]

        # Set a reasonable unit dimension
        scan_unit = crystal_map.scan_unit
        if scan_unit == "px":
            dim = "pixel-length"
        elif scan_unit[-1] == "m":
            dim = "si-length"  # Default
        else:
            dim = dimension._Dimension(scan_unit)

        # Set up arguments to AnchoredSizeBar() if not already present in kwargs
        d = dict(
            pad=0.2,
            sep=3,
            border_pad=0.5,
            location="lower left",
            box_alpha=0.6,
            dimension=dim,
        )
        [kwargs.setdefault(k, v) for k, v in d.items()]

        # Create scalebar
        bar = scalebar.ScaleBar(
            dx=crystal_map._step_sizes[horizontal],
            units=crystal_map.scan_unit,
            **kwargs,
        )
        self.axes.add_artist(bar)
        self.scalebar = bar

        return bar

    def add_overlay(self, crystal_map, item):
        """Use a crystal map property as gray scale values of a phase
        map.

        The property's range is adjusted to [0, 1] for maximum contrast.

        Parameters
        ----------
        crystal_map : orix.crystal_map.CrystalMap
            Crystal map object to obtain necessary data from.
        item : str
            Name of map property to scale phase array with. The property
            range is adjusted for maximum contrast.

        Examples
        --------
        >>> xmap
        Phase  Orientations   Name       Symmetry  Color
        1      5657 (48.4%)   austenite  432       tab:blue
        2      6043 (51.6%)   ferrite    432       tab:orange
        Properties: iq, dp
        Scan unit: um

        Plot a phase map with a map property as overlay

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(projection="plot_map")
        >>> im = ax.plot_map(xmap)
        >>> ax.add_overlay(xmap, xmap.dp)

        """
        image = self.images[0]
        image_data = image.get_array()

        if image_data.ndim < 3:
            # Adding overlay to a scalar plot (should this be allowed?)
            image_data = image.to_rgba(image_data)[:, :, :3]  # No alpha

        # Scale prop to [0, 1] to maximize image contrast
        overlay = crystal_map.get_map_data(item)
        overlay_min = np.nanmin(overlay)
        rescaled_overlay = (overlay - overlay_min) / (np.nanmax(overlay) - overlay_min)

        n_channels = 3
        for i in range(n_channels):
            image_data[:, :, i] *= rescaled_overlay

        image.set_data(image_data)

    def add_colorbar(self, label=None, **kwargs):
        """Add an opinionated colorbar to the figure and return it.

        The colorbar is also available as an attribute :attr:`colorbar`.

        Parameters
        ----------
        label : str, optional
            Colorbar title, default is ``None``.
        kwargs
            Keyword arguments passed to
            :meth:`mpl_toolkits.axes_grid1.make_axes_locatable.append_axes`.

        Returns
        -------
        cbar : matplotlib.colorbar
            Colorbar.

        Examples
        --------
        >>> xmap
        Phase  Orientations   Name       Symmetry  Color
        1      5657 (48.4%)   austenite  432       tab:blue
        2      6043 (51.6%)   ferrite    432       tab:orange
        Properties: iq, dp
        Scan unit: um

        Plot a map property and add a colorbar

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(projection="plot_map")
        >>> im = ax.plot_map(xmap, xmap.dp, cmap="inferno")
        >>> cbar = ax.add_colorbar("Dot product")

        If the default options are not satisfactory, the colorbar can be
        updated

        >>> cbar.ax.set_ylabel(ylabel="dp", rotation=90)
        """
        # Keyword arguments
        d = {"position": "right", "size": "5%", "pad": 0.1}
        [kwargs.setdefault(k, v) for k, v in d.items()]

        # Add colorbar
        divider = make_axes_locatable(self)
        cax = divider.append_axes(**kwargs)
        cbar = self.figure.colorbar(self.images[0], cax=cax)

        # Set label with padding
        cbar.ax.get_yaxis().labelpad = 15
        cbar.ax.set_ylabel(label, rotation=270)

        self.colorbar = cbar

        return cbar

    def remove_padding(self):
        """Remove all white padding outside of the figure.

        Examples
        --------
        >>> xmap
        Phase  Orientations   Name       Symmetry  Color
        1      5657 (48.4%)   austenite  432       tab:blue
        2      6043 (51.6%)   ferrite    432       tab:orange
        Properties: iq, dp
        Scan unit: um

        Remove all figure and axes padding of a phase map

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(projection="plot_map")
        >>> ax.plot_map(xmap)
        >>> ax.remove_padding()
        """
        self.set_axis_off()
        self.margins(0, 0)

        # Tune subplot layout
        cbar = self.images[0].colorbar
        if cbar is not None:
            right = self.figure.subplotpars.right
        else:
            right = 1
        self.figure.subplots_adjust(top=1, bottom=0, right=right, left=0)

    def _set_plot_shape(self, crystal_map, axes=None, depth=None):
        """Set `CrystalMapPlot` attributes describing which data axes to
        plot.

        Parameters
        ----------
        crystal_map : orix.crystal_map.CrystalMap
            Map to determine plotting axes and slices from.
        axes : list of ints, optional
            Data axes to plot. If ``None``, the last two data axes are
            plotted (default).
        depth : int, optional
            Which data layer to plot along the final axis not in `axes` if
            data is 3D. If ``None``, this is set to zero, i.e. the first
            layer (default).
        """
        ndim = crystal_map.ndim

        # Get data axes to plot
        if axes is None:
            axes = [ndim - 2, ndim - 1]
        axes = list(axes)
        axes.sort()
        self._data_axes = axes[:2]  # Can only plot two axes!

        if depth is None:  # Plot first layer
            depth = 0

        # Get data slices to plot
        slices = []
        data_shape = []
        for data_axis, axis_size in zip(
            crystal_map._coordinate_axes.keys(), crystal_map._original_shape
        ):
            data_slice = slice(depth, depth + 1, None)
            for plot_axis in self._data_axes:
                if data_axis == plot_axis:
                    data_slice = slice(None, None, None)
                    data_shape.append(axis_size)
            slices.append(data_slice)
        self._data_slices = tuple(slices)
        self._data_shape = tuple(data_shape)

    def _add_legend(self, patches, **kwargs):
        """Add a legend to the axes object.

        Parameters
        ----------
        patches : list of matplotlib.patches.Patch
            Patches with color code and name.
        kwargs
            Keyword arguments passed to :meth:`matplotlib.axes.legend`.
        """
        d = {
            "borderpad": 0.3,
            "handlelength": 0.75,
            "handletextpad": 0.3,
            "framealpha": 0.6,
        }
        [kwargs.setdefault(k, v) for k, v in d.items()]
        self.legend(handles=patches, **kwargs)

    def _override_status_bar(self, image, crystal_map):
        """Display coordinates, a property value (if scalar values are
        plotted), and Euler angles (in radians) per data point in the
        status bar.

        This is done by overriding
        :meth:`matplotlib.images.AxesImage.get_cursor_data`,
        :meth:`matplotlib.images.AxesImage.format_cursor_data` and
        :meth:`matplotlib.axes.Axes.format_coord`.

        Parameters
        ----------
        image : matplotlib.images.AxesImage
            Image object.
        crystal_map : orix.crystal_map.CrystalMap
            Crystal map object to obtain necessary data from.

        Returns
        -------
        image : matplotlib.images.AxesImage
            Image object where the above mentioned methods are overridden.
        """
        # Get data shape to plot
        n_rows, n_cols = self._data_shape

        # Get rotations, ensuring correct masking
        # TODO: Show orientations in Euler angles (computationally
        #  intensive...)
        r = crystal_map.get_map_data("rotations", decimals=3)
        r = r[self._data_slices].squeeze()

        # Get image data, overwriting potentially masked regions set to 0.0
        image_data = image.get_array()  # numpy.masked.MaskedArray
        # Force float because np.nan is a float
        image_data = image_data.astype("float")
        image_data[image_data.mask] = np.nan

        def status_bar_data(event):
            col = int(event.xdata + 0.5)
            row = int(event.ydata + 0.5)
            return row, col, r[row, col], image_data[row, col]

        # Set width of status bar fields
        x_width = len(str(n_cols - 1))
        y_width = len(str(n_rows - 1))
        scalar_width = len(str(np.nanmax(image_data)))

        # Override
        image.get_cursor_data = status_bar_data
        self.axes.format_coord = lambda x, y: ""

        def format_status_bar_data_rgb(data):
            """Status bar format for RGB plots."""
            return (
                f"(y,x):({data[0]:{y_width}},{data[1]:{x_width}})"
                f" rot:({data[2][0]:5},{data[2][1]:5},{data[2][2]:5})"
            )

        def format_status_bar_data_scalar(data):
            """Status bar format for scalar plots."""
            return (
                f"(y,x):({data[0]:{y_width}},{data[1]:{x_width}})"
                f" val:{data[3]:{scalar_width}}"
                f" rot:({data[2][0]:5},{data[2][1]:5},{data[2][2]:5})"
            )

        # Pick status bar format and override this as well
        if image_data.ndim > 2 and image_data.shape[-1] == 3:
            image.format_cursor_data = format_status_bar_data_rgb
        else:
            image.format_cursor_data = format_status_bar_data_scalar

        return image


register_projection(CrystalMapPlot)



__name__ = "orix"
__version__ = "0.8.1"
__author__ = "orix developers"
__author_email__ = "pyxem.team@gmail.com"
__description__ = "orix is an open-source python library for handling crystal orientation mapping data."
# Initial committer first, then sorted by line contributions
__credits__ = [
    "Ben Martineau",
    "Håkon Wiik Ånes",
    "Phillip Crout",
    "Duncan Johnstone",
    "Paddy Harrison",
    "Niels Cautaerts",
    "Simon Høgås",
]



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
from itertools import product

import numpy as np

from orix.scalar import Scalar
from orix.vector import Vector3d


class Miller(Vector3d):
    """Direct crystal lattice vectors (uvw or UVTW) and reciprocal
    crystal lattice vectors (hkl or hkil), the latter known as Miller
    indices, describing directions with respect to the crystal reference
    frame defined by a phase's crystal lattice and symmetry.
    """

    def __init__(self, xyz=None, uvw=None, UVTW=None, hkl=None, hkil=None, phase=None):
        r"""Create a set of direct lattice vectors (uvw or UVTW)
        or reciprocal lattice vectors (hkl or hkil) describing
        directions with respect to a crystal reference frame defined by
        a phase's crystal lattice and symmetry.

        Exactly one of `xyz`, `uvw`, `UVTW`, `hkl`, or `hkil` must be
        passed.

        The vectors are stored internally as cartesian coordinates in
        the `data` attribute.

        Parameters
        ----------
        xyz : numpy.ndarray, list, or tuple, optional
            Vector(s) given in cartesian coordinates. Default is None.
        uvw : numpy.ndarray, list, or tuple, optional
            Indices of direct lattice vector(s). Default is None.
        UVTW : numpy.ndarray, list, or tuple, optional
            Indices of direct lattice vector(s), often preferred over
            `uvw` in trigonal and hexagonal lattices. Default is None.
        hkl : numpy.ndarray, list, or tuple, optional
            Indices of reciprocal lattice vector(s). Default is None.
        hkil : numpy.ndarray, list, or tuple, optional
            Indices of reciprocal lattice vector(s), often preferred
            over `hkl` in trigonal and hexagonal lattices. Default is
            None.
        phase : orix.crystal_map.Phase, optional
            A phase with a crystal lattice and symmetry. Must be passed
            whenever direct or reciprocal lattice vectors are created.

        Notes
        -----
        The Miller-Bravais indices :math:`UVTW` are defined as

        .. math::
            U &= \frac{2u - v}{3}, \\
            V &= \frac{2v - u}{3}, \\
            T &= -\frac{u + v}{3}, \\
            W &= w.
        """
        n_passed = np.sum([i is not None for i in [xyz, uvw, UVTW, hkl, hkil]])
        if n_passed == 0 or n_passed > 1:
            raise ValueError(
                "Exactly one of `xyz`, `uvw`, `UVTW`, `hkl`, `hkil` must be passed"
            )
        if xyz is None and phase is None:
            raise ValueError(
                "A phase with a crystal lattice and symmetry must be passed to create "
                "direct or reciprocal lattice vector(s)"
            )

        self.phase = phase
        if xyz is not None:
            xyz = np.asarray(xyz)
            self.coordinate_format = "xyz"
        elif uvw is not None:
            xyz = _uvw2xyz(uvw=uvw, lattice=phase.structure.lattice)
            self.coordinate_format = "uvw"
        elif UVTW is not None:
            _check_UVTW(UVTW)
            uvw = _UVTW2uvw(UVTW=UVTW)
            xyz = _uvw2xyz(uvw=uvw, lattice=phase.structure.lattice)
            self.coordinate_format = "UVTW"
        elif hkl is not None:
            xyz = _hkl2xyz(hkl=hkl, lattice=phase.structure.lattice)
            self.coordinate_format = "hkl"
        elif hkil is not None:
            hkil = np.asarray(hkil)
            _check_hkil(hkil)
            hkl = _hkil2hkl(hkil)
            xyz = _hkl2xyz(hkl=hkl, lattice=phase.structure.lattice)
            self.coordinate_format = "hkil"
        super().__init__(xyz)

    def __repr__(self):
        """String representation."""
        name = self.__class__.__name__
        shape = self.shape
        symmetry = None if self.phase is None else self.phase.point_group.name
        coordinate_format = self.coordinate_format
        data = np.array_str(self.coordinates, precision=4, suppress_small=True)
        return (
            f"{name} {shape}, point group {symmetry}, {coordinate_format}\n" f"{data}"
        )

    def __getitem__(self, key):
        """NumPy fancy indexing of vectors."""
        m = self.__class__(xyz=self.data[key], phase=self.phase).deepcopy()
        m.coordinate_format = self.coordinate_format
        return m

    # ---------------------- Unique properties ---------------------- #

    @property
    def coordinate_format(self):
        """Vector coordinate format, either "xyz", "uvw", "UVTW", "hkl",
        or "hkil".
        """
        return self._coordinate_format

    @coordinate_format.setter
    def coordinate_format(self, value):
        """Set the vector coordinate format, either "xyz", "uvw",
        "UVTW", "hkl", or "hkil".
        """
        formats = ["xyz", "uvw", "UVTW", "hkl", "hkil"]
        if value not in formats:
            raise ValueError(f"Available coordinate formats are {formats}")
        self._coordinate_format = value

    @property
    def coordinates(self):
        """Vector coordinates."""
        coordinate_format = self.coordinate_format
        if coordinate_format == "xyz":
            coordinate_format = "data"
        coordinates = self.__getattribute__(coordinate_format)
        return coordinates

    @property
    def hkl(self):
        """Reciprocal lattice vectors."""
        return _xyz2hkl(xyz=self.data, lattice=self.phase.structure.lattice)

    @hkl.setter
    def hkl(self, value):
        """Set the reciprocal lattice vectors."""
        self.data = _hkl2xyz(hkl=value, lattice=self.phase.structure.lattice)

    @property
    def hkil(self):
        """Reciprocal lattice vectors expressed as 4-index
        Miller-Bravais indices.
        """
        return _hkl2hkil(self.hkl)

    @hkil.setter
    def hkil(self, value):
        r"""Set the reciprocal lattice vectors expressed as 4-index
        Miller-Bravais indices. The sum of the first three indices,
        :math:`h`, :math:`k`, and :math:`i` must be zero.
        """
        self.hkl = _hkil2hkl(value)

    @property
    def h(self):
        """First reciprocal lattice vector index."""
        return self.hkl[..., 0]

    @property
    def k(self):
        """Second reciprocal lattice vector index."""
        return self.hkl[..., 1]

    @property
    def i(self):
        r"""Third reciprocal lattice vector index in 4-index
        Miller-Bravais indices, equal to :math:`-(h + k)`.
        """
        return self.hkil[..., 2]

    @property
    def l(self):
        """Third reciprocal lattice vector index, or fourth index in
        4-index Miller Bravais indices.
        """
        return self.hkl[..., 2]

    @property
    def uvw(self):
        """Direct lattice vectors."""
        return _xyz2uvw(xyz=self.data, lattice=self.phase.structure.lattice)

    @uvw.setter
    def uvw(self, value):
        """Set the direct lattice vectors."""
        self.data = _uvw2xyz(uvw=value, lattice=self.phase.structure.lattice)

    @property
    def UVTW(self):
        r"""Direct lattice vectors expressed as 4-index Weber symbols.
        They are defined as

        .. math::
            U &= \frac{2u - v}{3}, \\
            V &= \frac{2v - u}{3}, \\
            T &= -\frac{u + v}{3}, \\
            W &= w.
        """
        return _uvw2UVTW(self.uvw)

    @UVTW.setter
    def UVTW(self, value):
        r"""Set the direct lattice vectors expressed as 4-index Weber
        symbols. The sum of the first three indices, :math:`U`,
        :math:`V`, and :math:`T` must be zero.
        """
        self.uvw = _UVTW2uvw(value)

    @property
    def u(self):
        """First direct lattice vector index."""
        return self.uvw[..., 0]

    @property
    def v(self):
        """Second direct lattice vector index."""
        return self.uvw[..., 1]

    @property
    def w(self):
        """Third direct lattice vector index."""
        return self.uvw[..., 2]

    @property
    def U(self):
        r"""First direct lattice vector index in 4-index Weber symbols,
        equal to :math:`(2u - v)/(3)`.
        """
        return self.UVTW[..., 0]

    @property
    def V(self):
        r"""Second direct lattice vector index in 4-index Weber symbols,
        equal to :math:`(2v - u)/3`.
        """
        return self.UVTW[..., 1]

    @property
    def T(self):
        r"""Third direct lattice vector index in 4-index Weber symbols,
        equal to :math:`-(u + v)/3`.
        """
        return self.UVTW[..., 2]

    @property
    def W(self):
        r"""Fourth direct lattice vector index in 4-index Weber symbols,
        equal to :math:`w`.
        """
        return self.UVTW[..., 3]

    @property
    def length(self):
        """Length of each vector given in lattice parameter units if
        the :attr:`coordinate_format` attribute equals "uvw" or "UVTW",
        and inverse lattice parameter units if the attribute equals
        "hkl" or "hkil". If the attribute equals "xyz", the norms of the
        vectors in :attr:`data` are returned.
        """
        if self.coordinate_format in ["hkl", "hkil"]:
            return self.phase.structure.lattice.rnorm(self.hkl)
        elif self.coordinate_format in ["uvw", "UVTW"]:
            return self.phase.structure.lattice.norm(self.uvw)
        else:
            return self.norm.data

    @property
    def multiplicity(self):
        """Number of symmetrically equivalent directions per vector."""
        l = self.symmetrise(unique=True, return_multiplicity=True)[1]
        return l.reshape(self.shape)

    @property
    def space(self):
        """Whether the vector is in "direct" or "reciprocal" space."""
        return dict(
            xyz="direct",
            uvw="direct",
            UVTW="direct",
            hkl="reciprocal",
            hkil="reciprocal",
        )[self.coordinate_format]

    @property
    def is_hexagonal(self):
        """Whether the crystal reference frame is hexagonal/trigonal or
        not.
        """
        return self.phase.is_hexagonal

    # ----------- Overwritten Vector3d/Object3d properties ----------- #

    @property
    def unit(self):
        """Unit vectors."""
        m = self.__class__(xyz=super().unit.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    # ------------------------ Unique methods ------------------------ #

    @classmethod
    def from_highest_indices(cls, phase, uvw=None, hkl=None):
        """Create a set of unique direct or reciprocal lattice vectors
        from three highest indices and a phase (crystal lattice and
        symmetry).

        Parameters
        ----------
        phase : orix.crystal_map.Phase
            A phase with a crystal lattice and symmetry.
        uvw : numpy.ndarray, list, or tuple, optional
            Three highest direct lattice vector indices.
        hkl : numpy.ndarray, list, or tuple, optional
            Three highest reciprocal lattice vector indices.
        """
        if uvw is not None:
            coordinate_format = "uvw"
            highest_idx = uvw
        elif hkl is not None:
            coordinate_format = "hkl"
            highest_idx = hkl
        else:
            raise ValueError("Either highest `hkl` or `uvw` indices must be passed")
        idx = _get_indices_from_highest(highest_indices=highest_idx)
        init_kw = {coordinate_format: idx, "phase": phase}
        return cls(**init_kw).unique()

    @classmethod
    def from_min_dspacing(cls, phase, min_dspacing=0.05):
        """Create a set of unique reciprocal lattice vectors with a
        a direct space interplanar spacing greater than a lower
        threshold.

        Parameters
        ----------
        phase : orix.crystal_map.Phase
            A phase with a crystal lattice and symmetry.
        min_dspacing : float, optional
            Smallest interplanar spacing to consider. Default is 0.05,
            in the unit used to define the lattice parameters in
            `phase`.
        """
        highest_hkl = _get_highest_hkl(
            lattice=phase.structure.lattice, min_dspacing=min_dspacing
        )
        hkl = _get_indices_from_highest(highest_indices=highest_hkl)
        return cls(hkl=hkl, phase=phase).unique()

    def deepcopy(self):
        """Return a deepcopy of the instance."""
        return deepcopy(self)

    def round(self, max_index=20):
        """Round a set of index triplet (Miller) or quartet
        (Miller-Bravais/Weber) to the *closest* smallest integers.

        Adopted from MTEX's Miller.round function.

        Parameters
        ----------
        max_index : int
            Maximum integer index to round to, by default 20.

        Return
        ------
        Miller
            Rounded set of index triplet(s) or quartet(s).
        """
        if self.coordinate_format == "xyz":
            return self.deepcopy()
        else:
            new_coords = _round_indices(indices=self.coordinates, max_index=max_index)
            init_kw = {self.coordinate_format: new_coords, "phase": self.phase}
            return self.__class__(**init_kw)

    def symmetrise(self, unique=False, return_multiplicity=False, return_index=False):
        """Vectors symmetrically equivalent to the ones in `self`.

        Parameters
        ----------
        unique : bool, optional
            Whether to return only unique vectors. Default is False.
        return_multiplicity : bool, optional
            Whether to return the multiplicity of each vector. Default
            is False.
        return_index : bool, optional
            Whether to return the index into `self` for the returned
            symmetrically equivalent vectors. Default is False.

        Returns
        -------
        Miller
            Flattened symmetrically equivalent vectors.
        multiplicity : numpy.ndarray
            Multiplicity of each vector. Returned if
            `return_multiplicity` is True.
        idx : numpy.ndarray
            Index into `self` for the symmetrically equivalent vectors.
            Returned if `return_index` is True.
        """
        if return_multiplicity and not unique:
            raise ValueError("`unique` must be True when `return_multiplicity` is True")
        elif return_index and not unique:
            raise ValueError("`unique` must be True when `return_index` is True")

        # Symmetrise directions with respect to crystal symmetry on the
        # flattened set of vectors
        operations = self.phase.point_group
        v2 = operations.outer(self.flatten())

        if unique:
            n_v = self.size  # Number of initial vectors in `self`

            # Array for symmetrically equivalent vectors
            v3 = self.zero((n_v, operations.size))

            # Array for multiplicity of initial vectors
            multiplicity = np.zeros(n_v, dtype=int)

            # Array for index into `self` for the returned symmetrically
            # equivalent vectors
            idx = np.ones(v3.size, dtype=int) * -1

            # Loop over initial vectors
            l_accum = 0
            for i in range(n_v):
                # Unique vectors among those symmetrically equivalent
                vi = v2[:, i].unique()
                l = vi.size  # Multiplicity
                v3[i, :l] = vi  # Insert only the unique ones

                # Multiplicity of this initial vector
                multiplicity[i] = l

                # Index into `self` for the unique, symmetrically
                # equivalent vectors
                idx[l_accum : l_accum + l] = i
                l_accum += l

            # Remove entries into `v3` and `idx` not used
            non_zero = np.sum(np.abs(v3.data), axis=-1) != 0
            v2 = v3[non_zero]
            idx = idx[: np.sum(non_zero)]

        v2 = v2.flatten()

        # Carry over crystal structure and coordinate format
        m = self.__class__(xyz=v2.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format

        if return_multiplicity and return_index:
            return m, multiplicity, idx
        elif return_multiplicity and not return_index:
            return m, multiplicity
        elif not return_multiplicity and return_index:
            return m, idx
        else:
            return m

    def _compatible_with(self, other, raise_error=False):
        """Whether `self` and `other` have both the same crystal lattice
        and symmetry and that the vectors are in the same space.

        Parameters
        ----------
        other : Miller
        raise_error : bool, optional
            Whether to raise a ValueError if the instances are not
            compatible. Default is False.

        Returns
        -------
        bool
        """
        same_symmetry = self.phase.point_group.name == other.phase.point_group.name
        same_lattice = np.allclose(
            self.phase.structure.lattice.abcABG(),
            other.phase.structure.lattice.abcABG(),
        )
        same_space = self.space == other.space
        compatible = same_symmetry * same_lattice * same_space
        if not compatible and raise_error:
            raise ValueError(
                "The crystal lattices and symmetries must be the same, and the vectors "
                "must be in the same space"
            )
        else:
            return compatible

    # ------------- Overwritten Vector3d/Object3d methods- ----------- #

    def angle_with(self, other, use_symmetry=False):
        """Calculate angles between vectors in `self` and `other`,
        possibly using symmetrically equivalent vectors to find the
        smallest angle under symmetry.

        Vectors must have compatible shapes, and be in the same space
        (direct or recprocal) and crystal reference frames.

        Parameters
        ----------
        other : Vector3d or Miller
        use_symmetry : bool, optional
            Whether to consider equivalent vectors to find the smallest
            angle under symmetry. Default is False.

        Returns
        -------
        Scalar
            The angle between the vectors, in radians.
        """
        self._compatible_with(other, raise_error=True)
        if use_symmetry:
            other2 = other.symmetrise(unique=True)
            cosines = self.dot_outer(other2).data / (
                self.norm.data[..., np.newaxis] * other2.norm.data[np.newaxis, ...]
            )
            cosines = np.round(cosines, 9)
            angles = np.min(np.arccos(cosines), axis=-1)
            return Scalar(angles)
        else:
            return super().angle_with(other)

    def cross(self, other):
        """Cross product of a direct or reciprocal lattice vector with
        another vector, which is considered the zone axis between the
        vectors.

        Vectors must have compatible shapes, and be in the same space
        (direct or recprocal) and crystal reference frames.

        Returns
        -------
        Miller
            Vectors in reciprocal (direct) space if direct (reciprocal)
            vectors are crossed.
        """
        self._compatible_with(other, raise_error=True)
        new_fmt = dict(hkl="uvw", uvw="hkl", hkil="UVTW", UVTW="hkil")
        m = self.__class__(xyz=super().cross(other).data, phase=self.phase)
        m.coordinate_format = new_fmt[self.coordinate_format]
        return m

    def dot(self, other):
        """Dot product of a vector with another vector.

        Vectors must have compatible shapes, and be in the same space
        (direct or recprocal) and crystal reference frames.

        Returns
        -------
        Scalar
        """
        self._compatible_with(other, raise_error=True)
        return super().dot(other)

    def dot_outer(self, other):
        """Outer dot product of a vector with another vector.

        Vectors must be in the same space (direct or recprocal) and
        crystal reference frames.

        The dot product for every combination of vectors in `self` and
        `other` is computed.

        Returns
        -------
        Scalar
        """
        self._compatible_with(other, raise_error=True)
        return super().dot_outer(other)

    def flatten(self):
        m = self.__class__(xyz=super().flatten().data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def transpose(self, *axes):
        """Returns a new Miller object containing the same data transposed.

        If ndim is originally 2, then order may be undefined.
        In this case the first two dimensions will be transposed.

        Parameters
        ----------
        axes : int, optional
            The transposed axes order. Only navigation axes need to be defined.
            May be undefined if self only contains two navigation dimensions.

        Returns
        -------
        Miller :
            A transposed Miller instance of the object.

        """
        m = self.__class__(xyz=super().transpose(*axes).data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def get_nearest(self):
        """NotImplemented."""
        return NotImplemented

    def mean(self, use_symmetry=False):
        """Mean vector of the set of vectors."""
        # TODO: Allow using symmetry by projecting to fundamental sector
        if use_symmetry:
            return NotImplemented
        m = self.__class__(xyz=super().mean().data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def reshape(self, *shape):
        m = self.__class__(xyz=super().reshape(*shape).data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def unique(self, use_symmetry=False, return_index=False):
        """Unique vectors in `self`.

        Parameters
        ----------
        use_symmetry : bool, optional
            Whether to consider equivalent vectors to compute the unique
            vectors. Default is False.
        return_index : bool, optional
            Whether to return the indices of the (flattened) data where
            the unique entries were found. Default is False.

        Returns
        -------
        Miller
            Flattened unique vectors.
        idx : numpy.ndarray
            Indices of the unique data in the (flattened) array.
        """
        out = super().unique(return_index=return_index)
        if return_index:
            v, idx = out
        else:
            v = out

        if use_symmetry:
            operations = self.phase.point_group
            n_v = v.size
            v2 = operations.outer(v).flatten()
            v2 = operations.outer(v).flatten().reshape(*(n_v, operations.size))
            data = v2.data
            data_sorted = np.zeros_like(data)
            for i in range(n_v):
                a = data[i]
                order = np.lexsort(a.T)  # Sort by column 1, 2, then 3
                data_sorted[i] = a[order]
            _, idx = np.unique(data_sorted, return_index=True, axis=0)
            v = v[idx[::-1]]

        m = self.__class__(xyz=v.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        if return_index:
            return m, idx
        else:
            return m


def _uvw2xyz(uvw, lattice):
    uvw = np.asarray(uvw)
    shape = uvw.shape
    uvw = uvw.reshape((uvw.size // 3, 3))
    xyz = _direct_structure_matrix(lattice).dot(uvw.T).T
    return xyz.reshape(shape)


def _xyz2uvw(xyz, lattice):
    xyz = np.asarray(xyz)
    shape = xyz.shape
    xyz = xyz.reshape((xyz.size // 3, 3))
    uvw = xyz.dot(_reciprocal_structure_matrix(lattice))
    return uvw.reshape(shape)


def _hkl2xyz(hkl, lattice):
    hkl = np.asarray(hkl)
    shape = hkl.shape
    hkl = hkl.reshape((hkl.size // 3, 3))
    xyz = _reciprocal_structure_matrix(lattice).dot(hkl.T).T
    return xyz.reshape(shape)


def _xyz2hkl(xyz, lattice):
    xyz = np.asarray(xyz)
    shape = xyz.shape
    xyz = xyz.reshape((xyz.size // 3, 3))
    hkl = xyz.dot(_direct_structure_matrix(lattice))
    return hkl.reshape(shape)


def _hkl2hkil(hkl):
    hkl = np.asarray(hkl)
    hkil = np.zeros(hkl.shape[:-1] + (4,))
    h = hkl[..., 0]
    k = hkl[..., 1]
    hkil[..., 0] = h
    hkil[..., 1] = k
    hkil[..., 2] = -(h + k)
    hkil[..., 3] = hkl[..., 2]
    return hkil


def _hkil2hkl(hkil):
    hkil = np.asarray(hkil)
    hkl = np.zeros(hkil.shape[:-1] + (3,))
    hkl[..., :2] = hkil[..., :2]
    hkl[..., 2] = hkil[..., 3]
    return hkl


def _check_hkil(hkil):
    hkil = np.asarray(hkil)
    if not np.allclose(np.sum(hkil[..., :3], axis=-1), 0, atol=1e-4):
        raise ValueError(
            "The Miller-Bravais indices convention h + k + i = 0 is not satisfied"
        )


def _uvw2UVTW(uvw, convention=None):
    uvw = np.asarray(uvw)
    UVTW = np.zeros(uvw.shape[:-1] + (4,))
    u = uvw[..., 0]
    v = uvw[..., 1]
    # DeGraef: U = (2u - v) / 3, V = (2v - u) / 3, T = -(u + v) / 3, W = w
    UVTW[..., 0] = (2 * u - v) / 3
    UVTW[..., 1] = (2 * v - u) / 3
    UVTW[..., 2] = -(u + v) / 3
    UVTW[..., 3] = uvw[..., 2]
    if convention is not None and convention.lower() == "mtex":
        # MTEX: U = 2u - v, V = 2v - u, T = -(u + v), W = 3w
        UVTW *= 3
    return UVTW


def _UVTW2uvw(UVTW, convention=None):
    UVTW = np.asarray(UVTW)
    uvw = np.zeros(UVTW.shape[:-1] + (3,))
    # DeGraef: u = 2U + V, v = 2V + U, w = W
    U = UVTW[..., 0]
    V = UVTW[..., 1]
    uvw[..., 0] = 2 * U + V
    uvw[..., 1] = U + 2 * V
    uvw[..., 2] = UVTW[..., 3]
    if convention is not None and convention.lower() == "mtex":
        # MTEX: u = 2U + V, v = 2V + U, w = W / 3
        uvw /= 3
    return uvw


def _check_UVTW(UVTW):
    UVTW = np.asarray(UVTW)
    if not np.allclose(np.sum(UVTW[..., :3], axis=-1), 0, atol=1e-4):
        raise ValueError(
            "The Miller-Bravais indices convention U + V + T = 0 is not satisfied"
        )


def _get_indices_from_highest(highest_indices):
    """Return a list of coordinates from a set of highest indices.

    Parameters
    ----------
    highest_indices : list, tuple, or numpy.ndarray
        Highest indices to consider.

    Returns
    -------
    indices : numpy.ndarray
        An array of indices sorted from positive to negative in the
        first column.
    """
    highest_indices = np.asarray(highest_indices)
    if not np.all(highest_indices >= 0) or np.all(highest_indices == 0):
        raise ValueError(
            f"All indices {highest_indices} must be positive with at least one non-zero"
        )
    index_ranges = [np.arange(-i, i + 1) for i in highest_indices]
    indices = np.asarray(list(product(*index_ranges)), dtype=int)
    indices = indices[~np.all(indices == 0, axis=1)]  # Remove (000)
    indices = indices[::-1]  # Make e.g. (111) first instead of (-1-1-1)
    return indices


def _get_highest_hkl(lattice, min_dspacing=0.05):
    """Return the highest Miller indices hkl of the plane with a direct
    space interplanar spacing closest to a threshold.

    Parameters
    ----------
    lattice : diffpy.structure.Lattice
        Crystal lattice.
    min_dspacing : float, optional
        Smallest interplanar spacing to consider. Default is 0.05.

    Returns
    -------
    highest_hkl : numpy.ndarray
        Highest Miller indices.
    """
    highest_hkl = np.ones(3, dtype=int)
    for i in range(3):
        hkl = np.zeros(3)
        hkl[i] = 1
        d = 1 / lattice.rnorm(hkl)
        while d > min_dspacing:
            hkl[i] += 1
            d = 1 / lattice.rnorm(hkl)
        highest_hkl[i] = hkl[i] - 1
    return highest_hkl


# TODO: Implement in diffpy.structure.Lattice
def _direct_structure_matrix(lattice):
    a, b, c = lattice.abcABG()[:3]
    ca, cb, cg = lattice.ca, lattice.cb, lattice.cg
    sg = lattice.sg
    return np.array(
        [
            [a, b * cg, c * cb],
            [0, b * sg, -c * (cb * cg - ca) / sg],
            [0, 0, lattice.volume / (a * b * sg)],
        ]
    )


# TODO: Implement in diffpy.structure.Lattice
def _reciprocal_structure_matrix(lattice):
    return np.linalg.inv(_direct_structure_matrix(lattice)).T


def _round_indices(indices, max_index=12):
    """Round a set of index triplet (Miller) or quartet (Miller-Bravais)
    to the *closest* smallest integers.

    Adopted from MTEX's Miller.round function.

    Parameters
    ----------
    indices : list, tuple, or numpy.ndarray
        Set of index triplet(s) or quartet(s) to round.
    max_index : int, optional
        Maximum integer index to round to, by default 12.

    Return
    ------
    new_indices : numpy.ndarray
        Integer array of rounded set of index triplet(s) or quartet(s).
    """
    # Allow list and tuple input (and don't overwrite `indices`)
    idx = np.asarray(indices)

    # Flatten and remove redundant third index if Miller-Bravais
    n_idx = idx.shape[-1]  # 3 or 4
    idx_flat = np.reshape(idx, (-1, n_idx))
    if n_idx == 4:
        idx_flat = idx_flat[..., [0, 1, 3]]

    # Get number of sets, max. index per set, and all possible integer
    # multipliers between 1 and `max_index`
    n_sets = idx_flat.size // 3
    max_per_set = np.max(np.abs(idx_flat), axis=-1)
    multipliers = np.arange(1, max_index + 1)

    # Divide by highest index, repeat array `max_index` number of times,
    # and multiply with all multipliers
    idx_scaled = (
        np.broadcast_to(idx_flat / max_per_set[..., np.newaxis], (max_index, n_sets, 3))
        * multipliers[..., np.newaxis, np.newaxis]
    )

    # Find the most suitable multiplier per set, which gives the
    # smallest error between the initial set and the scaled and rounded
    # set
    error = 1e-7 * np.round(
        1e7
        * np.sum((idx_scaled - np.round(idx_scaled)) ** 2, axis=-1)
        / np.sum(idx_scaled**2, axis=-1)
    )
    idx_min_error = np.argmin(error, axis=0)
    multiplier = (idx_min_error + 1) / max_per_set

    # Reshape `multiplier` to match indices shape
    multiplier = multiplier.reshape(idx.shape[:-1])[..., np.newaxis]

    # Finally, multiply each set with their most suitable multiplier,
    # and round
    new_indices = np.round(multiplier * idx).astype(int)

    return new_indices



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from h5py import File

from orix.io import save
from orix.io.plugins._h5ebsd import hdf5group2dict


class TestH5ebsd:
    def test_hdf5group2dict_update_dict(self, temp_file_path, crystal_map):
        """Can read datasets from an HDF5 file into an existing
        dictionary.
        """
        save(temp_file_path, crystal_map)
        with File(temp_file_path, mode="r") as f:
            this_dict = {"hello": "there"}
            this_dict = hdf5group2dict(f["crystal_map"], dictionary=this_dict)
            assert this_dict["hello"] == "there"
            assert this_dict["data"] == f["crystal_map/data"]
            assert this_dict["header"] == f["crystal_map/header"]



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Private tools for coloring crystal directions.

These functions are adopted from MTEX.
"""

import matplotlib.colors as mcolors
import numpy as np

from orix.quaternion import Rotation
from orix.vector import Vector3d


def polar_coordinates_in_sector(sector, v):
    r"""Calculate the polar coordinates of crystal direction(s)
    :class:`Vector3d` relative to the (bary)center of a
    :class:`FundamentalSector` of a Laue group.

    Parameters
    ----------
    sector : FundamentalSector
        A fundamental sector of a Laue group, described by a set of
        normal vectors, a center, and vertices.
    v : Vector3d
        Crystal direction(s) to get polar coordinates for.

    Returns
    -------
    azimuth : np.ndarray
        Azimuthal polar coordinate(s).
    polar : np.ndarray
        Polar polar coordinate(s).

    Notes
    -----
    This procedure is adopted from MTEX' :code:`polarCoordinates`
    function, which implements the coloring described in section 2.4 in
    :cite:`nolze2016orientation` (see Fig. 4 in that reference).

    The azimuthal coordinate is the angle to the barycenter relative to
    some fixed vertex of the sector, here chosen as the north pole
    [001]. The polar coordinate is the distance to the barycenter. The
    barycenter is the center of the fundamental sector.
    """
    center = sector.center.unit
    n_vertices = sector.vertices.size
    v = v.unit

    # Azimuthal coordinate
    if n_vertices == 0:
        # Point group Ci (-1) has no vertices
        rx = Vector3d.xvector() - center
    else:
        rx = Vector3d.zvector() - center  # North pole to sector center
    azimuth = _calculate_azimuth(center, rx, v)
    if n_vertices != 0:
        azimuth = _correct_azimuth(azimuth, sector, rx)

    # Polar coordinate
    if np.count_nonzero(sector.dot(center).data) == 0:
        polar = center.angle_with(v).data / np.pi
    else:
        # Normal to plane containing sector center and crystal direction
        v_center_normal = v.cross(center).unit
        polar = np.full(v.shape, np.inf)
        for normal in sector:
            boundary_points = v_center_normal.cross(normal).unit
            # Some boundary points are zero vectors
            with np.errstate(invalid="ignore"):
                distances_polar = (-v).angle_with(boundary_points).data
                distances_polar /= (-center).angle_with(boundary_points).data
            distances_polar[np.isnan(distances_polar)] = 1
            polar = np.minimum(polar, distances_polar)  # Element-wise

    return azimuth, polar


def rgb_from_polar_coordinates(azimuth, polar):
    """Calculate RGB colors from polar coordinates.

    Parameters
    ----------
    azimuth : np.ndarray
        Azimuthal coordinate(s).
    polar : np.ndarray
        Polar coordinate(s).

    Returns
    -------
    rgb : np.ndarray
        Color(s).
    """
    angle = np.mod(azimuth / (2 * np.pi), 1)
    h, s, v = hsl_to_hsv(angle, 1, polar)
    return mcolors.hsv_to_rgb(np.stack((h, s, v), axis=-1))


def hsl_to_hsv(hue, saturation, lightness):
    """Convert color described by HSL (hue, saturation and lightness) to
    HSV (hue, saturation and value).

    Adapted from MTEX' function :code:`hsl2hsv`.

    Parameters
    ----------
    hue : np.ndarray or float
        Hue(s). Not changed by the function, but included in input and
        output for convenience.
    saturation : np.ndarray or float
        Saturation value(s).
    lightness : np.ndarray or float
        Lightness value(s).

    Returns
    -------
    hue : np.ndarray or float
        The same copy of hue(s) as input.
    saturation2 : np.ndarray or float
        Adjusted saturation value(s).
    value : np.ndarray or float
        Value(s).
    """
    l2 = 2 * lightness
    s2 = saturation * np.where(l2 <= 1, l2, 2 - l2)
    saturation2 = (2 * s2) / (l2 + s2)
    saturation2[np.isnan(saturation2)] = 0
    value = (l2 + s2) / 2
    return hue, saturation2, value


def _calculate_azimuth(center, rx, v):
    """Calculate the azimuthal coordinate of the polar coordinates of a
    fundamental sector (inverse pole figure).

    Taken from MTEX' :code:`calcAngle` function.

    Parameters
    ----------
    center : ~orix.vector.Vector3d
    rx : ~orix.vector.Vector3d
    v : ~orix.vector.Vector3d

    Returns
    -------
    azimuth : numpy.ndarray
    """
    rx = (rx - rx.dot(center) * center).unit  # Orthogonal to center
    ry = center.cross(rx).unit  # Perpendicular to rx
    distances_azimuthal = (v - center).unit
    azimuth = np.arctan2(
        ry.dot(distances_azimuthal).data, rx.dot(distances_azimuthal).data
    )
    azimuth = np.mod(azimuth, 2 * np.pi)
    azimuth[np.isnan(azimuth)] = 0
    return azimuth


def _correct_azimuth(azimuth, sector, rx):
    """Correct the azimuthal coordinate of the polar coordinates of a
    fundamental sector (inverse pole figure) when vertices are not
    an equal distance from each other.

    This ensures for example that the entirely blue part of the m-3m
    sector is in the [111] direction, and not in the direction
    corresponding to the sector position perpendicular to the
    [001]-[101] line (straight north from the barycenter).

    Taken from MTEX' :code:`correctAngle` function.

    Parameters
    ----------
    azimuth : numpy.ndarray
    sector : ~orix.vector.FundamentalSector
    rx : ~orix.vector.Vector3d

    Returns
    -------
    azimuth : numpy.ndarray
        Corrected azimuthal coordinates.
    """
    center = sector.center.unit
    vertices = sector.vertices.unit

    m = 1000
    azimuth2 = np.linspace(0, 2 * np.pi, m)
    rot = Rotation.from_axes_angles(center, azimuth2[:-1])
    normals = rot * rx.cross(center).unit

    # Distance between center and all boundary points
    cross_outer = Vector3d.zero((sector.size, m - 1))
    for i in range(sector.size):
        cross_outer[i] = sector[i].cross(normals)
    polar = np.min(cross_outer.angle_with(center).data, axis=0)

    if vertices.size == 3:
        angle = _calculate_azimuth(center, rx, vertices)
        azimuth3 = np.round(m * np.sort(angle) / (2 * np.pi)).astype(int)
        if azimuth3[0] < 10:
            azimuth3 = np.delete(azimuth3, 0)
        idx = np.arange(azimuth3[0])
        polar[idx] /= np.sum(polar[idx]) / 3
        idx = np.arange(azimuth3[0], azimuth3[1])
        polar[idx] /= np.sum(polar[idx]) / 3
        idx = np.arange(azimuth3[1], polar.size)
        polar[idx] /= np.sum(polar[idx]) / 3

    polar = 2 * np.pi * np.cumsum(np.append(np.zeros(1), polar / np.sum(polar)))
    azimuth = np.interp(azimuth, azimuth2, polar)

    return azimuth



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from orix.crystal_map import CrystalMap, Phase
from orix.io import load, loadang, save
from orix.io.plugins.ang import (
    _get_header,
    _get_phases_from_header,
    _get_vendor_columns,
    _get_nrows_ncols_step_sizes,
    _get_column_width,
)
from orix.tests.conftest import (
    ANGFILE_TSL_HEADER,
    ANGFILE_ASTAR_HEADER,
    ANGFILE_EMSOFT_HEADER,
)


@pytest.mark.parametrize(
    "angfile_astar, expected_data",
    [
        (
            (
                (2, 5),
                (1, 1),
                np.ones(2 * 5, dtype=int),
                np.array(
                    [
                        [4.485496, 0.952426, 0.791507],
                        [1.343904, 0.276111, 0.825890],
                        [1.343904, 0.276111, 0.825890],
                        [1.343904, 0.276111, 0.825890],
                        [4.555309, 2.895152, 3.972020],
                        [1.361357, 0.276111, 0.825890],
                        [4.485496, 0.220784, 0.810182],
                        [0.959931, 2.369110, 4.058938],
                        [0.959931, 2.369110, 4.058938],
                        [4.485496, 0.220784, 0.810182],
                    ],
                ),
            ),
            np.array(
                [
                    [0.77861956, -0.12501022, 0.44104243, 0.42849224],
                    [0.46256046, -0.13302712, -0.03524667, -0.87584204],
                    [0.46256046, -0.13302712, -0.03524667, -0.87584204],
                    [0.46256046, -0.13302712, -0.03524667, -0.87584204],
                    [0.05331986, 0.95051048, 0.28534763, -0.11074093],
                    [0.45489991, -0.13271448, -0.03640618, -0.87984517],
                    [0.8752001, -0.02905178, 0.10626836, 0.47104969],
                    [0.3039118, 0.01972273, -0.92612154, 0.22259272],
                    [0.3039118, 0.01972273, -0.92612154, 0.22259272],
                    [0.8752001, -0.02905178, 0.10626836, 0.47104969],
                ]
            ),
        ),
    ],
    indirect=["angfile_astar"],
)
def test_loadang(angfile_astar, expected_data):
    loaded_data = loadang(angfile_astar)
    assert np.allclose(loaded_data.data, expected_data)


class TestAngReader:
    @pytest.mark.parametrize(
        "angfile_tsl, map_shape, step_sizes, phase_id, example_rot",
        [
            (
                # Read by angfile_tsl() via request.param (passed via `indirect` below)
                (
                    (5, 3),  # map_shape
                    (0.1, 0.1),  # step_sizes
                    np.zeros(5 * 3, dtype=int),  # phase_id
                    5,  # n_unknown_columns
                    np.array(
                        [[1.59942, 2.37748, 4.53419], [1.59331, 2.37417, 4.53628]]
                    ),  # rotations as rows of Euler angle triplets
                ),
                (5, 3),
                (0.1, 0.1),
                np.zeros(5 * 3, dtype=int),
                np.array(
                    [[1.59942, 2.37748, 4.53419], [1.59331, 2.37417, 4.53628]]
                ),  # rotations as rows of Euler angle triplets
            ),
            (
                (
                    (8, 4),  # map_shape
                    (1.5, 1.5),  # step_sizes
                    np.zeros(8 * 4, dtype=int),  # phase_id
                    5,  # n_unknown_columns
                    np.array(
                        [[5.81107, 2.34188, 4.47345], [6.16205, 0.79936, 1.31702]]
                    ),  # rotations as rows of Euler angle triplets
                ),
                (8, 4),
                (1.5, 1.5),
                np.zeros(8 * 4, dtype=int),
                np.array(
                    [[5.81107, 2.34188, 4.47345], [6.16205, 0.79936, 1.31702]]
                ),  # rotations as rows of Euler angle triplets
            ),
        ],
        indirect=["angfile_tsl"],
    )
    def test_load_ang_tsl(
        self, angfile_tsl, map_shape, step_sizes, phase_id, example_rot
    ):
        xmap = load(angfile_tsl)

        # Fraction of non-indexed points
        non_indexed_fraction = int(np.prod(map_shape) * 0.1)
        assert non_indexed_fraction == np.sum(~xmap.is_indexed)

        # Properties
        assert list(xmap.prop.keys()) == [
            "iq",
            "ci",
            "unknown1",
            "fit",
            "unknown2",
            "unknown3",
            "unknown4",
            "unknown5",
        ]

        # Coordinates
        ny, nx = map_shape
        dy, dx = step_sizes
        assert np.allclose(xmap.x, np.tile(np.arange(nx) * dx, ny))
        assert np.allclose(xmap.y, np.sort(np.tile(np.arange(ny) * dy, nx)))

        # Map shape and size
        assert xmap.shape == map_shape
        assert xmap.size == np.prod(map_shape)

        # Attributes are within expected ranges or have a certain value
        assert xmap.prop["ci"].max() <= 1
        assert xmap["indexed"].fit.max() <= 3
        assert all(xmap["not_indexed"].fit == 180)
        assert all(xmap["not_indexed"].ci == -1)

        # Phase IDs (accounting for non-indexed points)
        phase_id[xmap["not_indexed"].id] = -1
        assert np.allclose(xmap.phase_id, phase_id)

        # Rotations
        rot_unique = np.unique(xmap["indexed"].rotations.to_euler(), axis=0)
        assert np.allclose(
            np.sort(rot_unique, axis=0), np.sort(example_rot, axis=0), atol=1e-5
        )
        assert np.allclose(
            xmap["not_indexed"].rotations.to_euler()[0],
            np.array([np.pi, 0, np.pi]),
            atol=1e-5,
        )

        # Phases
        assert xmap.phases.size == 2  # Including non-indexed
        assert xmap.phases.ids == [-1, 0]
        phase = xmap.phases[0]
        assert phase.name == "Aluminum"
        assert phase.point_group.name == "432"

    @pytest.mark.parametrize(
        "angfile_astar, map_shape, step_sizes, phase_id, example_rot",
        [
            (
                # Read by angfile_astar() via request.param (passed via `indirect`
                # below)
                (
                    (9, 3),  # map_shape
                    (4.5, 4.5),  # step_sizes
                    np.ones(9 * 3, dtype=int),  # phase_id
                    np.array(
                        [
                            [1.895079, 0.739496, 1.413542],
                            [1.897871, 0.742638, 1.413717],
                        ]
                    ),
                ),
                (9, 3),
                (4.5, 4.5),
                np.ones(9 * 3, dtype=int),
                np.array(
                    [[1.895079, 0.739496, 1.413542], [1.897871, 0.742638, 1.413717]]
                ),
            ),
            (
                (
                    (11, 13),  # map_shape
                    (10, 10),  # step_sizes
                    np.ones(11 * 13, dtype=int),  # phase_id
                    np.array(
                        [
                            [1.621760, 2.368935, 4.559324],
                            [1.604481, 2.367539, 4.541870],
                        ]
                    ),
                ),
                (11, 13),
                (10, 10),
                np.ones(11 * 13, dtype=int),
                np.array(
                    [[1.621760, 2.368935, 4.559324], [1.604481, 2.367539, 4.541870]]
                ),
            ),
        ],
        indirect=["angfile_astar"],
    )
    def test_load_ang_astar(
        self, angfile_astar, map_shape, step_sizes, phase_id, example_rot
    ):
        cm = load(angfile_astar)

        # Properties
        assert list(cm.prop.keys()) == ["ind", "rel", "relx100"]

        # Coordinates
        ny, nx = map_shape
        dy, dx = step_sizes
        assert np.allclose(cm.x, np.tile(np.arange(nx) * dx, ny))
        assert np.allclose(cm.y, np.sort(np.tile(np.arange(ny) * dy, nx)))

        # Map shape and size
        assert cm.shape == map_shape
        assert cm.size == np.prod(map_shape)

        # Attributes are within expected ranges or have a certain value
        assert cm.prop["ind"].max() <= 100
        assert cm.prop["rel"].max() <= 1
        assert cm.prop["relx100"].max() <= 100
        relx100 = (cm.prop["rel"] * 100).astype(int)
        assert np.allclose(cm.prop["relx100"], relx100)

        # Phase IDs
        assert np.allclose(cm.phase_id, phase_id)

        # Rotations
        rot_unique = np.unique(cm.rotations.to_euler(), axis=0)
        assert np.allclose(
            np.sort(rot_unique, axis=0), np.sort(example_rot, axis=0), atol=1e-6
        )

        # Phases
        assert cm.phases.size == 1
        assert cm.phases.ids == [1]
        phase = cm.phases[1]
        assert phase.name == "Nickel"
        assert phase.point_group.name == "432"

    @pytest.mark.parametrize(
        "angfile_emsoft, map_shape, step_sizes, phase_id, example_rot",
        [
            (
                # Read by angfile_emsoft() via request.param (passed via `indirect`
                # below)
                (
                    (10, 11),  # map_shape
                    (4.5, 4.5),  # step_sizes
                    np.concatenate(
                        (
                            np.ones(int(np.ceil((10 * 11) / 2))),
                            np.ones(int(np.floor((10 * 11) / 2))) * 2,
                        )
                    ),  # phase_id
                    np.array(
                        [
                            [1.895079, 0.739496, 1.413542],
                            [1.897871, 0.742638, 1.413717],
                        ]
                    ),
                ),
                (10, 11),
                (4.5, 4.5),
                np.concatenate(
                    (
                        np.ones(int(np.ceil((10 * 11) / 2))),
                        np.ones(int(np.floor((10 * 11) / 2))) * 2,
                    )
                ),
                np.array(
                    [[1.895079, 0.739496, 1.413542], [1.897871, 0.742638, 1.413717]]
                ),
            ),
            (
                (
                    (3, 6),  # map_shape
                    (10, 10),  # step_sizes
                    np.concatenate(
                        (
                            np.ones(int(np.ceil((3 * 6) / 2))),
                            np.ones(int(np.floor((3 * 6) / 2))) * 2,
                        )
                    ),  # phase_id
                    np.array(
                        [[1.62176, 2.36894, 1.72386], [1.60448, 2.36754, 1.72386]]
                    ),
                ),
                (3, 6),
                (10, 10),
                np.concatenate(
                    (
                        np.ones(int(np.ceil((3 * 6) / 2))),
                        np.ones(int(np.floor((3 * 6) / 2))) * 2,
                    )
                ),
                np.array([[1.62176, 2.36894, 1.72386], [1.60448, 2.36754, 1.72386]]),
            ),
        ],
        indirect=["angfile_emsoft"],
    )
    def test_load_ang_emsoft(
        self, angfile_emsoft, map_shape, step_sizes, phase_id, example_rot
    ):
        cm = load(angfile_emsoft)

        # Properties
        assert list(cm.prop.keys()) == ["iq", "dp"]

        # Coordinates
        ny, nx = map_shape
        dy, dx = step_sizes
        assert np.allclose(cm.x, np.tile(np.arange(nx) * dx, ny))
        assert np.allclose(cm.y, np.sort(np.tile(np.arange(ny) * dy, nx)))

        # Map shape and size
        assert cm.shape == map_shape
        assert cm.size == np.prod(map_shape)

        # Attributes are within expected ranges or have a certain value
        assert cm.prop["iq"].max() <= 100
        assert cm.prop["dp"].max() <= 1

        # Phase IDs
        assert np.allclose(cm.phase_id, phase_id)

        # Rotations
        rot_unique = np.unique(cm.rotations.to_euler(), axis=0)
        assert np.allclose(
            np.sort(rot_unique, axis=0), np.sort(example_rot, axis=0), atol=1e-5
        )

        # Phases (change if file header is changed!)
        phases_in_data = cm["indexed"].phases_in_data
        assert phases_in_data.size == 2
        assert phases_in_data.ids == [1, 2]
        assert phases_in_data.names == ["austenite", "ferrite/ferrite"]
        assert [i.name for i in phases_in_data.point_groups] == ["432"] * 2

    def test_get_header(self, temp_ang_file):
        temp_ang_file.write(ANGFILE_ASTAR_HEADER)
        temp_ang_file.close()
        assert _get_header(open(temp_ang_file.name)) == [
            "# File created from ACOM RES results",
            "# ni-dislocations.res",
            "#     ".rstrip(),
            "#     ".rstrip(),
            "# MaterialName      Nickel",
            "# Formula",
            "# Symmetry          43",
            "# LatticeConstants  3.520  3.520  3.520  90.000  90.000  90.000",
            "# NumberFamilies    4",
            "# hklFamilies       1  1  1 1 0.000000",
            "# hklFamilies       2  0  0 1 0.000000",
            "# hklFamilies       2  2  0 1 0.000000",
            "# hklFamilies       3  1  1 1 0.000000",
            "#",
            "# GRID: SqrGrid#",
        ]

    @pytest.mark.parametrize(
        "expected_vendor, expected_columns, vendor_header",
        [
            (
                "tsl",
                [
                    "iq",
                    "ci",
                    "phase_id",
                    "unknown1",
                    "fit",
                    "unknown2",
                    "unknown3",
                    "unknown4",
                    "unknown5",
                ],
                ANGFILE_TSL_HEADER,
            ),
            ("astar", ["ind", "rel", "phase_id", "relx100"], ANGFILE_ASTAR_HEADER),
            ("emsoft", ["iq", "dp", "phase_id"], ANGFILE_EMSOFT_HEADER),
        ],
    )
    def test_get_vendor_columns(
        self, expected_vendor, expected_columns, vendor_header, temp_ang_file
    ):
        expected_columns = ["euler1", "euler2", "euler3", "x", "y"] + expected_columns
        n_cols_file = len(expected_columns)

        temp_ang_file.write(vendor_header)
        temp_ang_file.close()
        header = _get_header(open(temp_ang_file.name))
        vendor, column_names = _get_vendor_columns(header, n_cols_file)

        assert vendor == expected_vendor
        assert column_names == expected_columns

    @pytest.mark.parametrize("n_cols_file", [15, 20])
    def test_get_vendor_columns_unknown(self, temp_ang_file, n_cols_file):
        temp_ang_file.write("Look at me!\nI'm Mr. .ang file!\n")
        temp_ang_file.close()
        header = _get_header(open(temp_ang_file.name))
        with pytest.warns(UserWarning, match=f"Number of columns, {n_cols_file}, "):
            vendor, column_names = _get_vendor_columns(header, n_cols_file)
            assert vendor == "unknown"
            expected_columns = [
                "euler1",
                "euler2",
                "euler3",
                "x",
                "y",
                "unknown1",
                "unknown2",
                "phase_id",
            ] + ["unknown" + str(i + 3) for i in range(n_cols_file - 8)]
            assert column_names == expected_columns

    @pytest.mark.parametrize(
        "header_phase_part, expected_names, expected_point_groups, "
        "expected_lattice_constants, expected_phase_id",
        [
            (
                [
                    [
                        "# Phase 42",
                        "# MaterialName      Nickel",
                        "# Formula",
                        "# Symmetry          43",
                        "# LatticeConstants  3.520  3.520  3.520  90.000  90.000  "
                        "90.000",
                    ],
                    [
                        "# MaterialName      Aluminium",
                        "# Formula  Al",
                        "# Symmetry          m3m",
                        "# LatticeConstants  3.520  3.520  3.520  90.000  90.000  "
                        "90.000",
                    ],
                ],
                ["Nickel", "Aluminium"],
                ["43", "m3m"],
                [[3.52, 3.52, 3.52, 90, 90, 90], [3.52, 3.52, 3.52, 90, 90, 90]],
                [42, 43],
            ),
        ],
    )
    def test_get_phases_from_header(
        self,
        header_phase_part,
        expected_names,
        expected_point_groups,
        expected_lattice_constants,
        expected_phase_id,
    ):
        # Create header from parts
        header = [
            "# File created from ACOM RES results",
            "# ni-dislocations.res",
            "#     ",
            "#     ",
        ]
        hkl_families = [
            "# NumberFamilies    4",
            "# hklFamilies       1  1  1 1 0.000000",
            "# hklFamilies       2  0  0 1 0.000000",
            "# hklFamilies       2  2  0 1 0.000000",
            "# hklFamilies       3  1  1 1 0.000000",
        ]
        for phase in header_phase_part:
            header += phase + hkl_families
        header += [
            "#",
            "# GRID: SqrGrid#",
        ]
        ids, names, point_groups, lattice_constants = _get_phases_from_header(header)

        assert names == expected_names
        assert point_groups == expected_point_groups
        assert np.allclose(lattice_constants, expected_lattice_constants)
        assert np.allclose(ids, expected_phase_id)


class TestAngWriter:
    def test_write_read_loop(self, crystal_map, tmp_path):
        fname = tmp_path / "test_write_read_loop.ang"
        save(filename=fname, object2write=crystal_map)
        xmap_reload = load(filename=fname)

        assert np.allclose(
            xmap_reload.rotations.to_euler(), crystal_map.rotations.to_euler()
        )
        assert np.allclose(xmap_reload.phase_id - 1, crystal_map.phase_id)

    @pytest.mark.parametrize(
        "crystal_map_input, desired_shape, desired_step_sizes",
        [
            (((1, 4, 3), (1, 2, 3), 1, [0]), (4, 3), (2, 3)),
            (((1, 1, 3), (1, 1, 1), 1, [0]), (1, 3), (1, 1)),
            (((1, 1, 6), (1, 1, 3.14), 1, [0]), (1, 6), (1, 3.14)),
        ],
        indirect=["crystal_map_input"],
    )
    def test_get_nrows_ncols_step_sizes(
        self, crystal_map_input, desired_shape, desired_step_sizes
    ):
        xmap = CrystalMap(**crystal_map_input)
        nrows, ncols, dy, dx = _get_nrows_ncols_step_sizes(xmap)

        assert (nrows, ncols) == desired_shape
        assert (dy, dx) == desired_step_sizes

    @pytest.mark.parametrize(
        "max_value, decimals, expected_width",
        [(3.14, 2, 5), (3.1415, 2, 5), (3.141592, 6, 9)],
    )
    def test_get_column_width(self, max_value, decimals, expected_width):
        assert _get_column_width(max_value, decimals) == expected_width

    @pytest.mark.parametrize(
        "extra_prop", ["a", ["abc", "iq"], ["scores", "simulation_indices"]]
    )
    def test_extra_prop(self, crystal_map, tmp_path, extra_prop):
        fname = tmp_path / "test_extra_prop.ang"
        desired_arrays = []
        n_points = crystal_map.size
        for i, name in enumerate(extra_prop):
            new_array = np.arange(n_points) * i
            crystal_map.prop[name] = new_array
            desired_arrays.append(new_array)

        save(fname, crystal_map, extra_prop=extra_prop)
        xmap_reload = load(fname)

        for name in extra_prop:
            assert np.allclose(xmap_reload.prop[name], crystal_map.prop[name])

    def test_non_indexed_points(self, crystal_map, tmp_path):
        crystal_map[2].phase_id = -1
        fname = tmp_path / "test_non_indexed_points.ang"
        save(fname, crystal_map)

        xmap_reload = load(fname)

        crystal_map.phases[0].name = "phase1"
        assert xmap_reload.phases.names == crystal_map.phases.names

        new_phase_ids = xmap_reload.phase_id
        new_phase_ids[xmap_reload.is_indexed] -= 1
        assert np.allclose(new_phase_ids, crystal_map.phase_id)

    def test_points_not_in_data(self, crystal_map, tmp_path):
        crystal_map.prop["iq"] = np.ones(crystal_map.size)
        crystal_map[2].iq = 0
        xmap2 = crystal_map[crystal_map.iq == 1]
        fname = tmp_path / "test_points_not_in_data.ang"
        save(fname, xmap2)

        xmap_reload = load(fname)
        assert not xmap2.is_in_data.all()
        assert np.allclose(xmap2.is_in_data, xmap_reload.is_indexed)

    @pytest.mark.parametrize(
        "extra_phase_names", ["ni", ["ferrite", "austenite"], ["si", "al", "cu"]]
    )
    def test_extra_phases(self, crystal_map, tmp_path, extra_phase_names):
        crystal_map.phases.add_not_indexed()
        for i, name in enumerate(extra_phase_names):
            crystal_map.phases.add(Phase(name=name))
            crystal_map[i].phase_id = crystal_map.phases.id_from_name(name)
        fname = tmp_path / "test_extra_phases.ang"
        save(fname, crystal_map)
        xmap_reload = load(fname)

        crystal_map.phases[0].name = "phase1"
        assert np.allclose(xmap_reload.phase_id - 1, crystal_map.phase_id)

        pl = crystal_map.phases
        del pl[-1]
        assert xmap_reload.phases.names == pl.names

    @pytest.mark.parametrize("point_group", ["432", "121", "222"])
    def test_point_group_aliases(self, crystal_map, tmp_path, point_group):
        crystal_map.phases[0].point_group = point_group
        fname = tmp_path / "test_point_group_aliases.ang"
        save(fname, crystal_map)
        xmap_reload = load(fname)

        phase_ids = xmap_reload.phases.ids
        assert xmap_reload.phases[phase_ids[0]].point_group.name == point_group

    @pytest.mark.parametrize(
        "crystal_map_input",
        [((1, 1, 5), (1, 1, 2), 1, [0])],
        indirect=["crystal_map_input"],
    )
    def test_1d_map(self, crystal_map_input, tmp_path):
        xmap = CrystalMap(**crystal_map_input)
        assert xmap.ndim == 1
        fname = tmp_path / "test_1d_map.ang"
        save(fname, xmap)

        xmap_reload = load(fname)

        assert xmap_reload.ndim == 1
        assert np.allclose(xmap.rotations.to_euler(), xmap_reload.rotations.to_euler())

    @pytest.mark.parametrize(
        "crystal_map_input",
        [((3, 3, 3), (1, 2, 3), 1, [0])],
        indirect=["crystal_map_input"],
    )
    def test_3d_map_raises(self, crystal_map_input, tmp_path):
        xmap = CrystalMap(**crystal_map_input)
        fname = tmp_path / "test_3d_raises.ang"
        with pytest.raises(ValueError, match="Writing a 3D dataset to an .ang file"):
            save(fname, xmap)

    @pytest.mark.parametrize(
        "crystal_map_input, index",
        [
            (((1, 4, 3), (1, 2, 3), 5, [0]), 0),
            (((1, 4, 3), (1, 2, 3), 5, [0]), 1),
            (((1, 4, 3), (1, 2, 3), 5, [0]), 4),
        ],
        indirect=["crystal_map_input"],
    )
    def test_write_data_index(self, crystal_map_input, tmp_path, index):
        xmap = CrystalMap(**crystal_map_input)
        xmap.prop["ci"] = np.arange(xmap.size * xmap.rotations_per_point).reshape(
            (xmap.size, xmap.rotations_per_point)
        )
        xmap.prop["iq"] = np.arange(xmap.size)
        extra_prop = "iq_times_ci"
        xmap.prop[extra_prop] = xmap.ci * xmap.iq[:, np.newaxis]
        fname = tmp_path / "test_write_data_layer_i.ang"
        save(fname, xmap, index=index, extra_prop=[extra_prop, "iq"])

        xmap_reload = load(fname)

        assert np.allclose(
            xmap.rotations[:, index].to_euler(), xmap_reload.rotations.to_euler()
        )
        assert np.allclose(xmap_reload.iq, xmap.iq)
        assert np.allclose(xmap_reload.ci, xmap.ci[:, index])
        assert np.allclose(xmap_reload.iq_times_ci, xmap.iq_times_ci[:, index])



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Stereographic projection of 3D vectors."""

import numpy as np

from orix.vector import SphericalRegion, Vector3d


_UPPER_HEMISPHERE = SphericalRegion([0, 0, 1])
_LOWER_HEMISPHERE = SphericalRegion([0, 0, -1])


class StereographicProjection:
    """Get stereographic coordinates (X, Y) from representations of
    vectors.
    """

    def __init__(self, pole=-1):
        """Initialize projection by setting whether the south pole (-1)
        or north pole (1) is the projection point.

        Parameters
        ----------
        pole : int, optional
            -1 or 1, where -1 (1) means the projection point is the
            south (north) pole [00-1] ([001]), i.e. only vectors with
            z > 0 (z < 0) are returned.
        """
        self.pole = pole
        self.region = SphericalRegion([0, 0, pole * -1])

    def vector2xy(self, v):
        r"""Return stereographic coordinates (X, Y) of 3D unit vectors.

        Parameters
        ----------
        v : Vector3d
            If it's not a unit vector, it will be made into one.

        Returns
        -------
        x : numpy.ndarray
            Stereographic coordinate X of shape same shape as the input
            vector shape. Only the vectors with :math:`z` coordinate
            positive (`pole` = -1) or negative (`pole` = 1) are
            returned.
        y : numpy.ndarray
            Stereographic coordinate Y of shape same shape as the input
            vector shape. Only the vectors with :math:`z` coordinate
            positive (`pole` = -1) or negative (`pole` = 1) are
            returned.

        Notes
        -----
        The stereographic coordinates :math:`(X, Y)` are calculated from
        the unit vectors' cartesian coordinates :math:`(x, y, z)` as

        .. math::
            (X, Y) = \left(\frac{-px}{z - p}, \frac{-py}{z - p}\right),

        where :math:`p` is either 1 (north pole as projection point) or
        -1 (south pole as projection point).
        """
        v = v[v <= self.region]
        return _vector2xy(v, pole=self.pole)

    def spherical2xy(self, azimuth, polar):
        r"""Return stereographic coordinates (X, Y) from 3D unit vectors
        created from spherical coordinates, azimuth :math:`\phi` and
        polar :math:`\theta`, defined as in the ISO 31-11 standard
        [SphericalWolfram]_.

        Parameters
        ----------
        azimuth : float or numpy.ndarray
            Spherical azimuth coordinate.
        polar : float or numpy.ndarray
            Spherical polar coordinate.

        Returns
        -------
        x : numpy.ndarray
            Stereographic coordinate X of shape same shape as the input
            vector shape. Only the vectors with :math:`z` coordinate
            positive (`pole` = -1) or negative (`pole` = 1) are
            returned.
        y : numpy.ndarray
            Stereographic coordinate Y of shape same shape as the input
            vector shape. Only the vectors with :math:`z` coordinate
            positive (`pole` = -1) or negative (`pole` = 1) are
            returned.

        See Also
        --------
        vector2xy
        """
        v = Vector3d.from_polar(azimuth=azimuth, polar=polar)
        return self.vector2xy(v)

    @staticmethod
    def vector2xy_split(v):
        """Return two sets of stereographic coordinates (X, Y) from 3D
        unit vectors: one set for vectors in the upper hemisphere, and
        one for the lower.

        Parameters
        ----------
        v : Vector3d
            If it's not a unit vector, it will be made into one.

        Returns
        -------
        x_upper : numpy.ndarray
            Stereographic coordinate X of upper hemisphere vectors, of
            shape same shape as the input vector shape.
        y_upper : numpy.ndarray
            Stereographic coordinate Y of upper hemisphere vectors, of
            shape same shape as the input vector shape.
        x_lower : numpy.ndarray
            Stereographic coordinate X of lower hemisphere vectors, of
            shape same shape as the input vector shape.
        y_lower : numpy.ndarray
            Stereographic coordinate Y of lower hemisphere vectors, of
            shape same shape as the input vector shape.

        See Also
        --------
        vector2xy
        """
        x_upper, y_upper = _vector2xy(v[v <= _UPPER_HEMISPHERE], pole=-1)
        x_lower, y_lower = _vector2xy(v[v <= _LOWER_HEMISPHERE], pole=1)
        return x_upper, y_upper, x_lower, y_lower

    def spherical2xy_split(self, azimuth, polar):
        r"""Return two sets of stereographic coordinates (X, Y) from 3D
        unit vectors created from spherical coordinates, azimuth
        :math:`\phi` and polar :math:`\theta`, defined as in the
        ISO 31-11 standard [SphericalWolfram]_: one set for vectors in
        the upper hemisphere, and one for the lower.

        Parameters
        ----------
        azimuth : float or numpy.ndarray
            Spherical azimuth coordinate.
        polar : float or numpy.ndarray
            Spherical polar coordinate.

        Returns
        -------
        x_upper : numpy.ndarray
            Stereographic coordinate X of upper hemisphere vectors, of
            shape same shape as the input vector shape.
        y_upper : numpy.ndarray
            Stereographic coordinate Y of upper hemisphere vectors, of
            shape same shape as the input vector shape.
        x_lower : numpy.ndarray
            Stereographic coordinate X of lower hemisphere vectors, of
            shape same shape as the input vector shape.
        y_lower : numpy.ndarray
            Stereographic coordinate Y of lower hemisphere vectors, of
            shape same shape as the input vector shape.

        See Also
        --------
        vector2xy
        """
        v = Vector3d.from_polar(azimuth=azimuth, polar=polar)
        return self.vector2xy_split(v)

    @property
    def inverse(self):
        """Return the corresponding inverse projection,
        :class:`InverseStereographicProjection`, with the same
        projection pole.
        """
        return InverseStereographicProjection(pole=self.pole)


def _vector2xy(v, pole):
    """Return stereographic coordinates (X, Y) of 3D unit vectors.

    (X, Y) is both zero for vectors with z equal to the projection pole.

    Parameters
    ----------
    v : Vector3d
        If it's not a unit vector, it will be made into one.
    pole : int
        -1 or 1, where -1 (1) means the projection point of the
        stereographic transform is the south (north) pole [00-1]
        ([001]), i.e. only vectors with z > 0 (z < 0) are returned.

    Returns
    -------
    x : numpy.ndarray
        Stereographic coordinate X of shape same shape as the input
        vector shape.
    y : numpy.ndarray
        Stereographic coordinate Y of shape same shape as the input
        vector shape.

    See Also
    --------
    StereographicProjection.vector2xy
    """
    vx, vy, vz = v.unit.xyz
    # We explicitly say (X, Y) is zero when the denominator is zero
    denom = vz - pole
    not_zero = denom != 0
    x = np.divide(-pole * vx, denom, where=not_zero, out=np.zeros_like(vx))
    y = np.divide(-pole * vy, denom, where=not_zero, out=np.zeros_like(vy))
    return x, y


class InverseStereographicProjection:
    """Get 3D unit vectors or spherical coordinates from stereographic
    coordinates (X, Y).
    """

    def __init__(self, pole=-1):
        """Initialize inverse projection by setting whether the south
        pole (-1) or north pole (1) is the projection point.

        Parameters
        ----------
        pole : int, optional
            -1 or 1, where -1 (1) means the projection point is the
            south (north) pole [00-1] ([001]), i.e. only vectors with
            z > 0 (z < 0) are returned.
        """
        self.pole = pole

    def xy2vector(self, x, y):
        r"""Return 3D unit vectors from stereographic coordinates
        (X, Y).

        Parameters
        ----------
        x : float or numpy.ndarray
        y : float or numpy.ndarray

        Returns
        -------
        Vector3d
            Unit vectors corresponding to the stereographic coordinates.
            Whether the upper or lower hemisphere points are returned is
            controlled by `pole` (-1 = upper, 1 = lower).

        Notes
        -----
        The 3D unit vectors :math:`(x, y, z)` are calculated from the
        stereographic coordinates :math:`(X, Y)` as

        .. math::
            (x, y, z) = \left(
                \frac{2x}{1 + x^2 + y^2},
                \frac{2y}{1 + x^2 + y^2},
                \frac{-p(1 - x^2 - y^2)}{1 + x^2 + y^2}
            \right),

        where :math:`p` is either 1 (north pole as projection point) or
        -1 (south pole as projection point).
        """
        denom = 1 + x**2 + y**2
        vx = 2 * x / denom
        vy = 2 * y / denom
        vz = -self.pole * (1 - x**2 - y**2) / denom
        return Vector3d(np.column_stack([vx, vy, vz]))

    def xy2spherical(self, x, y):
        r"""Return spherical coordinates, azimuth :math:`phi` and
        polar :math:`theta`, defined as in the ISO 31-11 standard
        [SphericalWolfram]_, from stereographic coordinates (X, Y).

        Parameters
        ----------
        x : float or numpy.ndarray
        y : float or numpy.ndarray

        Returns
        -------
        azimuth : numpy.ndarray
            Azimuth spherical coordinate corresponding to (X, Y).
            Whether the coordinates for the upper or lower hemisphere
            points are returned is controlled by `pole` (-1 = upper,
            1 = lower).
        theta : numpy.ndarray
            Polar spherical coordinate corresponding to (X, Y). Whether
            the coordinates for the upper or lower hemisphere points are
            returned is controlled by `pole` (-1 = upper, 1 = lower).

        See Also
        --------
        xy2vector
        StereographicProjection.spherical2xy
        """
        v = self.xy2vector(x=x, y=y)
        return v.azimuth.data, v.polar.data



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Reader of a crystal map from Bruker's h5ebsd file format."""

from diffpy.structure import Atom, Lattice, Structure
import numpy as np

from orix.crystal_map import CrystalMap, Phase, PhaseList
from orix.io.plugins._h5ebsd import H5ebsdFile
from orix.quaternion import Rotation


__all__ = ["file_reader"]

# Plugin description
format_name = "bruker_h5ebsd"
manufacturer = "Bruker"
file_extensions = ["h5", "hdf5", "h5ebsd"]
writes = False
writes_this = CrystalMap


def file_reader(filename, **kwargs):
    """Return a :class:`~orix.crystal_map.CrystalMap` from a file in
    Bruker Nano's dot product file format.

    Parameters
    ----------
    filename : str
        Path and file name.
    kwargs
        Keyword arguments passed to :func:`h5py.File`.

    Returns
    -------
    CrystalMap
    """
    f = BrukerH5ebsdFile(filename)
    f.open(**kwargs)
    f.set_scan_group_names()
    f.set_sem_group_file_location()
    f.read_data_into_dictionaries()
    f.close()
    f.set_map_shape()  # Necessary when checking if we can read the data
    if not f.can_read():
        raise ValueError(
            "Cannot return a crystal map from the file data because only a rectangular"
            "region of interest is supported"
        )
    f.set_crystal_map_data()
    return f.get_crystal_map()


class BrukerH5ebsdFile(H5ebsdFile):
    """Bruker Nano's HDF5 file in the h5ebsd format containing
    orientation data from Hough indexing, to be returned as a crystal
    map.
    """

    dont_read_in_data = ["RawPatterns"]
    dont_read_in_header = [
        "CameraTilt",
        "Coordinate Systems",
        "DetectorFullHeightMicrons",
        "DetectorFullWidthMicrons",
        "KV",
        "MADMax",
        "Magnification",
        "MapStepFactor",
        "MaxRadonBandCount",
        "MinIndexedBands",
        "NPoints",
        "OriginalFile",
        "PatternHeight",
        "PatternWidth",
        "PixelByteCount",
        "SEPixelSizeX",
        "SEPixelSizeY",
        "SampleTilt",
        "TopClip",
        "UnClippedPatternHeight",
        "WD",
        "XSTEP",
        "YSTE",
        "ZOffset",
    ]
    dont_read_in_sem = ["ZOffset"]
    is_rectangular = True
    map_cols = None
    map_rows = None
    scan_unit = "um"

    def can_read(self):
        """Return whether the file can be read."""
        square_grid = self.header_dict["Grid Type"] == "isometric"
        return self.is_rectangular * square_grid

    def final_preparations(self):
        """Final preparations of data before creation of a crystal map."""
        if self.map_rows is not None and self.map_cols is not None:
            # Sort data points into correct order
            rc = np.array([self.map_rows, self.map_cols])
            map_order = np.ravel_multi_index(rc, self.map_shape).argsort()
            self.x = self.x[map_order]
            self.phase_id = self.phase_id[map_order]
            self.rotations = self.rotations[map_order]
            for key, value in self.properties.items():
                self.properties[key] = value[map_order]
        self.x = self.x[::-1]

    def read_data_into_dictionaries(self):
        """Read data from the HDF5 file into dictionaries."""
        if self.sem_group_location is not None:
            self.sem_dict = self.get_dictionary(
                self.sem_group_location, recursive=True, dont_read=self.dont_read_in_sem
            )
        eg_name = self.scan_groups[0] + "/EBSD/"
        self.header_dict = self.get_dictionary(
            eg_name + "Header", recursive=True, dont_read=self.dont_read_in_header
        )
        self.data_dict = self.get_dictionary(
            eg_name + "Data", recursive=True, dont_read=self.dont_read_in_data
        )

    def set_sem_group_file_location(self):
        """Set 'SEM' group HDF5 file location. This can either be
        'Scan 1/SEM' or 'Scan 1/EBSD/SEM'.
        """
        sg = self.scan_groups[0]
        potential_places = [sg, sg + "/EBSD"]
        location = None
        for pp in potential_places:
            if "SEM" in self.file[pp].keys():
                location = pp + "/SEM"
        self.sem_group_location = location

    def set_coordinate_arrays(self):
        """Set coordinate arrays from dictionaries."""
        y = self.properties["YSAMPLE"]
        x = self.properties["XSAMPLE"]
        self.y = y - np.min(y)
        self.x = x - np.min(x)

    def set_crystal_map_data(self):
        """Set necessary crystal map data from dictionaries."""
        self.set_properties()
        self.set_coordinate_arrays()
        self.set_phase_id()
        self.set_phase_list()
        self.set_rotations()
        self.final_preparations()

    def set_map_shape(self):
        """Set the number of map rows and columns.

        Also try to set the map row and column position of each point in
        the data arrays, if the data sets 'Scan 1/SEM/IY' and
        'Scan 1/SEM/IX' are present. If not, it is assumed that the
        order of the data points is correct, and can be reshaped into a
        2D map without changing the order.
        """
        sd = self.sem_dict
        potential_names_y = ["IY", "SEM IY"]
        potential_names_x = ["IX", "SEM IX"]
        match_y = None
        match_x = None
        for key in sd.keys():
            if key in potential_names_y:
                match_y = key
            elif key in potential_names_x:
                match_x = key
        if match_y is not None and match_x is not None:
            map_rows = self.sem_dict[match_y]
            map_cols = self.sem_dict[match_x]

            # If False, we cannot read the data
            self.is_rectangular = _roi_is_rectangular(map_rows, map_cols)

            min_r, max_r = np.min(map_rows), np.max(map_rows)
            min_c, max_c = np.min(map_cols), np.max(map_cols)
            nrows = max_r - min_r + 1
            ncols = max_c - min_c + 1
            self.map_rows = map_rows - min_r
            self.map_cols = map_cols - min_c
        else:
            nrows = self.header_dict["NROWS"]
            ncols = self.header_dict["NCOLS"]
        self.map_shape = (nrows, ncols)

    def set_phase_id(self):
        """Set phase ID array from dictionaries."""
        self.phase_id = self.data_dict["Phase"]

    def set_phase_list(self):
        """Set phase list from dictionaries."""
        phase_list = dict2phaselist(self.header_dict["Phases"])
        phase_id = self.phase_id
        if 0 in phase_id:
            phase_list.add_not_indexed()
            phase_id[phase_id == 0] = -1
        self.phase_id = phase_id
        self.phase_list = phase_list

    def set_properties(self):
        """Set dictionary of property arrays from dictionaries."""
        self.properties = dict(
            PCX=self.data_dict["PCX"],
            PCY=self.data_dict["PCY"],
            DD=self.data_dict["DD"],
            MAD=self.data_dict["MAD"],
            MADPhase=self.data_dict["MADPhase"],
            NIndexedBands=self.data_dict["NIndexedBands"],
            RadonBandCount=self.data_dict["RadonBandCount"],
            RadonQuality=self.data_dict["RadonQuality"],
            XBEAM=self.data_dict["X BEAM"],
            YBEAM=self.data_dict["Y BEAM"],
            XSAMPLE=self.data_dict["X SAMPLE"],
            YSAMPLE=self.data_dict["Y SAMPLE"],
            ZSAMPLE=self.data_dict["Z SAMPLE"],
        )

    def set_rotations(self):
        """Set rotations from dictionaries."""
        dd = self.data_dict
        euler = np.column_stack([dd["phi1"], dd["PHI"], dd["phi2"]])
        euler = np.deg2rad(euler)
        self.rotations = Rotation.from_euler(euler)


def _roi_is_rectangular(map_rows, map_cols):
    """Return whether points in a map from Bruker Nano's h5ebsd file
    are in a rectangle.

    Parameters
    ----------
    map_rows : numpy.ndarray
    map_cols : numpy.ndarray

    Returns
    -------
    bool
    """
    map_rows_unique, map_rows_unique_counts = np.unique(map_rows, return_counts=True)
    map_cols_unique, map_cols_unique_counts = np.unique(map_cols, return_counts=True)
    return (
        np.all(np.diff(np.sort(map_rows_unique)) == 1)
        and np.all(np.diff(np.sort(map_cols_unique)) == 1)
        and np.unique(map_rows_unique_counts).size == 1
        and np.unique(map_cols_unique_counts).size == 1
    )


def dict2phaselist(dictionary):
    """Return a list of phases from a dictionary with keys and values
    from a Bruker Nano h5ebsd file.

    Parameters
    ----------
    dictionary : dict

    Returns
    -------
    PhaseList
    """
    return PhaseList(phases={int(k): dict2phase(v) for k, v in dictionary.items()})


def dict2phase(dictionary):
    """Return a phase from a dictionary with keys and values from a
    Bruker Nano h5ebsd file.

    Parameters
    ----------
    dictionary : dict

    Returns
    -------
    Phase
    """
    lattice_dict = dict(
        zip(["a", "b", "c", "alpha", "beta", "gamma"], dictionary["LatticeConstants"])
    )
    lattice = Lattice(**lattice_dict)
    atoms = [str2atom(atom) for atom in dictionary["AtomPositions"].values()]
    structure = Structure(lattice=lattice, atoms=atoms)
    structure.title = dictionary["Name"]
    return Phase(
        name=dictionary["Name"], space_group=int(dictionary["IT"]), structure=structure
    )


def str2atom(atom_positions):
    """Return an atom from a string in the format used by Bruker Nano
    in their h5ebsd file.

    Parameters
    ----------
    atom_positions : str

    Returns
    -------
    diffpy.structure.Atom
    """
    atom_positions = atom_positions.split(",")
    return Atom(
        atype=atom_positions[0],
        xyz=np.array(atom_positions[1:4]),
        occupancy=int(atom_positions[-1]),
    )



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.
import itertools

import numpy as np
import pytest

from orix.vector import AxAngle, Vector3d
from orix.quaternion import Rotation


axes = [
    (1, 0, 0),
    (1, 1, 0),
    (2, -1, 0),
    (0, 2, 1),
    (-1, -1, -1),
]

angles = [
    -2 * np.pi,
    -5 * np.pi / 6,
    -np.pi / 3,
    0,
    np.pi / 12,
    np.pi / 3,
    3 * np.pi / 4,
    2 * np.pi,
    np.pi / 7,
]

axangles = [np.array(angle) * Vector3d(axis).unit for axis in axes for angle in angles]
axangles += [
    np.array(angle) * Vector3d(axis).unit
    for axis in itertools.combinations_with_replacement(axes, 2)
    for angle in itertools.combinations_with_replacement(angles, 2)
]


@pytest.fixture(params=axangles[:100])
def axangle(request):
    return AxAngle(request.param.data)


def test_angle(axangle):
    assert np.allclose(axangle.angle.data, axangle.norm.data)


def test_axis(axangle):
    assert axangle.axis.shape == axangle.shape


@pytest.mark.parametrize(
    "axis, angle, expected_axis",
    [
        ((2, 1, 1), np.pi / 4, (0.816496, 0.408248, 0.408248)),
        (Vector3d((2, 0, 0)), -2 * np.pi, (-1, 0, 0)),
    ],
)
def test_from_axes_angles(axis, angle, expected_axis):
    ax = AxAngle.from_axes_angles(axis, angle)
    assert np.allclose(ax.axis.data, expected_axis)
    assert np.allclose(ax.angle.data, abs(angle))


@pytest.mark.parametrize(
    "rotation, expected",
    [(Rotation([1, 0, 0, 0]), [0, 0, 0]), (Rotation([0, 1, 0, 0]), [np.pi, 0, 0])],
)
def test_from_rotation(rotation, expected):
    axangle = AxAngle.from_rotation(rotation)
    assert np.allclose(axangle.data, expected)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""The fundamental sector for a symmetry in the stereographic
projection.
"""

import numpy as np

from orix.vector import SphericalRegion, Vector3d


_EDGE_STEPS = 1000


class FundamentalSector(SphericalRegion):
    """Fundamental sector for a symmetry in the inverse pole figure,
    defined by a set of sector normals.
    """

    # This is only set for T (23), Th (m-3) and O (432), in the
    # Symmetry.fundamental_sector property, because the UV S2 sampling
    # isn't uniform enough to produce the correct center according to
    # MTEX
    _center = None

    # Used when sorting `edges` for restricting stereographic plot
    _pole = -1

    @property
    def vertices(self):
        n = self.size
        if n == 0:
            return Vector3d.empty()
        else:
            normals = self.reshape(1, n)
            u = normals.cross(normals.transpose())
            return Vector3d(u[u <= self]).unique().unit

    @property
    def center(self):
        """Center vector of the fundamental sector.

        Taken from MTEX' :code:`sphericalRegion.center`.
        """
        v = self.vertices.unique()
        n_vertices = v.size
        n_normals = self.size
        if n_normals < 2:
            center = self
        elif n_vertices < 3:
            # Find the pair of maximum angle
            angles = self.angle_with(self.reshape(n_normals, 1)).data
            indices = np.argmax(angles, axis=1)
            center = self[indices].mean()
        elif n_vertices < 4:
            center = v.mean()
        elif isinstance(self._center, Vector3d):
            # Only the case for T (23), Th (m-3) and O (432), for which
            # the S2 sampling isn't uniform enough to produce the
            # correct center according to MTEX
            center = self._center
        else:
            # Avoid circular import
            from orix.sampling import sample_S2_cube_mesh

            v_all = sample_S2_cube_mesh(resolution=1)
            v = v_all[v_all < self]
            center = v.mean()

        return Vector3d(center)

    @property
    def edges(self):
        """Unit vectors which delineates the region in the stereographic
        projection.

        They are sorted in the counter-clockwise direction around the
        sector center in the stereographic projection.

        The first edge is repeated at the end. This is done so that
        :meth:`orix.plot.StereographicPlot.plot` draws bounding lines
        without gaps.
        """
        if self.size == 0:
            return Vector3d.empty()

        circles = self.get_circle(steps=_EDGE_STEPS)
        edges = np.zeros((self.size * _EDGE_STEPS + 3, 3))
        vertices = self.vertices

        if vertices.size == 0:
            return circles.squeeze()

        j = 0
        for ci, vi in zip(circles, vertices):
            # Only get the parts of the great circle that are within
            # this spherical region
            ci_inside = ci[ci <= self]
            v_keep = Vector3d(np.vstack((ci_inside.data, vi.data)))
            v_keep = v_keep.unique()
            order = np.lexsort((v_keep.azimuth.data, v_keep.polar.data))
            v_keep = v_keep[order]
            v_n = v_keep.size
            edges[j : j + v_n] = v_keep.data
            j += v_n
        edges = Vector3d(edges[:j])

        order = _order_to_sort_around_center(edges, self.center, self._pole)
        sorted_edges = edges[order]

        return sorted_edges.squeeze()


def _order_to_sort_around_center(v, center, pole=-1):
    vz = Vector3d.zvector()
    if pole == 1:
        vz = -vz
    angle = vz.angle_with(center).data
    axis = vz.cross(center)
    v_rotated = v.rotate(axis=axis, angle=-angle)

    order1 = np.argsort(v_rotated.azimuth.data)
    idx_closest_to_001 = np.argmax(v[order1].dot(vz).data)
    order2 = np.roll(order1, shift=-idx_closest_to_001)

    return order2


def _closed_edges_in_hemisphere(edges, sector, pole=-1):
    if pole == -1:
        is_outside = edges.polar.data >= np.pi / 2
    else:  # pole == 1
        is_outside = edges.polar.data <= np.pi / 2

    if not np.any(is_outside):
        return edges
    elif np.all(is_outside):
        return Vector3d.empty()
    else:
        idx_after_crossing_equator = np.where(is_outside != is_outside[0])[0][0]
        equator = Vector3d.zvector().get_circle(steps=_EDGE_STEPS)
        equator_inside = equator[equator <= sector]
        edges_inside = edges[~is_outside]
        azimuth_before_equator = edges_inside[
            [idx_after_crossing_equator - 1, idx_after_crossing_equator]
        ].azimuth.data
        v_before_equator = Vector3d.from_polar(
            azimuth_before_equator, polar=[np.pi / 2] * 2
        )
        return Vector3d(
            np.vstack(
                (
                    edges_inside[:idx_after_crossing_equator].data,
                    v_before_equator[0].data,
                    equator_inside.data,
                    v_before_equator[1].data,
                    edges_inside[idx_after_crossing_equator:].data,
                )
            )
        )



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import numpy as np

from orix.plot import EulerColorKey, IPFColorKeyTSL
from orix.quaternion import Orientation, symmetry
from orix.vector import Vector3d


class TestIPFColorKeyTSL:
    def test_orientation2color(self):
        # Color vertices of Oh IPF red, green and blue
        pg_o = symmetry.O  # 432
        pg_oh = pg_o.laue  # m-3m
        ori = Orientation.from_euler(
            np.radians(((0, 0, 0), (0, 45, 0), (-45, 54.7356, 45))),
            symmetry=pg_oh,
        )
        ckey_oh = IPFColorKeyTSL(pg_o)
        assert np.allclose(ckey_oh.symmetry.data, pg_oh.data)
        assert np.allclose(ckey_oh.direction.data, (0, 0, 1))
        assert repr(ckey_oh) == "IPFColorKeyTSL, symmetry: m-3m, direction: [0 0 1]"
        fig_o = ckey_oh.plot(return_figure=True)
        ax_o = fig_o.axes[0]
        assert ax_o._symmetry.name == pg_oh.name
        rgb_oh = ckey_oh.orientation2color(ori)
        assert np.allclose(rgb_oh, ((1, 0, 0), (0, 1, 0), (0, 0, 1)), atol=0.1)

        # Color [001] and "diagonals" of 2/m IPF red, green and blue
        pg_c2 = symmetry.C2  # 2
        pg_c2h = pg_c2.laue  # 2/m
        ori2 = Orientation.from_euler(
            np.radians(((-90, -90, 0), (0, 90, -55), (0, 90, 55))),
            symmetry=pg_c2h,
        )
        ckey_c2h = IPFColorKeyTSL(pg_c2, Vector3d.xvector())
        assert np.allclose(ckey_c2h.symmetry.data, pg_c2h.data)
        assert np.allclose(ckey_c2h.direction.data, (1, 0, 0))
        assert repr(ckey_c2h) == "IPFColorKeyTSL, symmetry: 2/m, direction: [1 0 0]"
        rgb_c2h = ckey_c2h.orientation2color(ori2)
        assert np.allclose(rgb_c2h, ((1, 0, 0), (0, 1, 0.23), (0, 0.23, 1)), atol=0.2)

        # Color vertices of D3d IPF red, green and blue
        pg_d3d = symmetry.D3d  # -3m
        ori3 = Orientation.from_euler(
            np.radians(((0, 0, 0), (0, -90, 60), (0, 90, 60))),
            symmetry=pg_d3d,
        )
        ckey_d3d = IPFColorKeyTSL(pg_d3d)
        rgb_d3d = ckey_d3d.orientation2color(ori3)
        assert np.allclose(rgb_d3d, ((1, 0, 0), (0, 1, 0), (0, 0, 1)), atol=1e-2)

    def test_triclinic(self):
        # Complete circle, three vectors on equator 120 degrees apart
        pg_c1 = symmetry.C1
        ori = Orientation.from_euler(
            np.radians(((0, 90, 90), (0, 90, -30), (0, -90, 30))), symmetry=pg_c1
        )
        ckey_c1 = IPFColorKeyTSL(pg_c1)
        rgb = ckey_c1.orientation2color(ori)
        assert np.allclose(rgb, ((1, 0, 0), (0, 1, 0), (0, 0, 1)))


class TestEulerColorKey:
    def test_orientation2color(self):
        # (2 pi, pi, 2 pi) and some random orientations
        ori = Orientation(
            (
                (0, -1, 0, 0),
                (0.4094, 0.7317, -0.4631, -0.2875),
                (-0.3885, 0.5175, -0.7589, 0.0726),
                (-0.5407, -0.7796, 0.2955, -0.1118),
                (-0.3874, 0.6708, -0.1986, 0.6004),
            )
        )

        ckey_1 = EulerColorKey(symmetry.C1)
        assert repr(ckey_1) == (
            "EulerColorKey, symmetry 1\n" "Max (phi1, Phi, phi2): (360, 180, 360)"
        )
        rgb_1 = ckey_1.orientation2color(ori)
        assert np.allclose(
            rgb_1,
            (
                (0, 1, 0),
                (0.508, 0.666, 0.687),
                (0.875, 0.741, 0.184),
                (0.410, 0.628, 0.525),
                (0.113, 0.493, 0.205),
            ),
            atol=1e-3,
        )

        ckey_432 = EulerColorKey(symmetry.O)
        assert repr(ckey_432) == (
            "EulerColorKey, symmetry 432\n" "Max (phi1, Phi, phi2): (360, 90, 90)"
        )
        rgb_432 = ckey_432.orientation2color(ori)
        assert np.allclose(
            rgb_432,
            (
                (0, 1, 0),
                (0.508, 1, 1),
                (0.875, 1, 0.737),
                (0.410, 1, 1),
                (0.113, 0.987, 0.818),
            ),
            atol=1e-3,
        )

    def test_plot(self):
        ckey_432 = EulerColorKey(symmetry.O)
        fig = ckey_432.plot(return_figure=True)
        assert isinstance(fig, plt.Figure)
        assert len(fig.axes) == 3

        labels = ["$\\phi_1$", "$\\Phi$", "$\\phi_2$"]
        angles = ckey_432.symmetry.euler_fundamental_region
        for i, ax in enumerate(fig.axes):
            if i == 0:
                assert ax.get_title() == "432"
            texts = ax.texts
            assert texts[0].get_text() == labels[i]
            assert texts[1].get_text() == str(angles[i]) + "$^{\\circ}$"

        plt.close("all")



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from orix.plot import IPFColorKeyTSL
from orix.plot.direction_color_keys._util import polar_coordinates_in_sector
from orix.quaternion import symmetry
from orix.vector import Vector3d


class TestDirectionColorKeyTSL:
    def test_direction2color(self):
        ckey_oh = IPFColorKeyTSL(symmetry.Oh)
        ckey_oh_direction = ckey_oh.direction_color_key
        assert repr(ckey_oh_direction) == "DirectionColorKeyTSL, symmetry m-3m"

    def test_triclinic(self):
        # Get RGB colors for C1. Will never reach from IPFColorKeyTSL
        # since Ci (-1) is the Laue group of C1.
        sector = symmetry.C1.fundamental_sector
        rgb2 = polar_coordinates_in_sector(sector, Vector3d.xvector())
        assert rgb2[0].size == rgb2[1].size == 0



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from datetime import datetime
import inspect
import os
from os.path import relpath, dirname
import re
import sys

from orix import __author__, __version__

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath("."))
sys.path.append("../")

project = "orix"
copyright = f"2018-{str(datetime.now().year)}, {__author__}"
author = __author__
release = __version__


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinxcontrib.bibtex",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
    "sphinx_gallery.load_style",
    "nbsphinx",
]

# Create links to references within orix's documentation to these packages
intersphinx_mapping = {
    "dask": ("https://docs.dask.org/en/latest", None),
    "diffpy.structure": ("https://www.diffpy.org/diffpy.structure", None),
    "h5py": ("http://docs.h5py.org/en/stable/", None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Syntax highlighting
pygments_style = "friendly"

# Logo
html_logo = "_static/img/orix_logo.png"
html_favicon = "_static/img/orix_logo.png"

# nbsphinx configuration
# Taken from nbsphinx' own nbsphinx configuration file, with slight
# modification to point nbviewer and Binder to the GitHub master links
# when the documentation is launched from a orix version with "dev" in
# the version
if "dev" in __version__:
    release_version = "master"
else:
    release_version = "v" + __version__
# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = (
    r"""
{% set docname = 'doc/' + env.doc2path(env.docname, base=None) %}

.. raw:: html

    <style>a:hover { text-decoration: underline; }</style>

    <div class="admonition note">
      This page was generated from
      <a class="reference external" href="https://github.com/pyxem/orix/blob/"""
    + f"{release_version}"
    + r"""/{{ docname|e }}">{{ docname|e }}</a>.
      Interactive online version:
      <span style="white-space: nowrap;"><a href="https://mybinder.org/v2/gh/pyxem/orix/"""
    + f"{release_version}"
    + r"""?filepath={{ docname|e }}"><img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a>.</span>
      <script>
        if (document.location.host) {
          $(document.currentScript).replaceWith(
            '<a class="reference external" ' +
            'href="https://nbviewer.jupyter.org/url' +
            (window.location.protocol == 'https:' ? 's/' : '/') +
            window.location.host +
            window.location.pathname.slice(0, -4) +
            'ipynb">View in <em>nbviewer</em></a>.'
          );
        }
      </script>
    </div>

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from
    \sphinxcode{\sphinxupquote{\strut {{ docname | escape_latex }}}} \dotfill}}
"""
)
# https://nbsphinx.readthedocs.io/en/0.8.0/never-execute.html
nbsphinx_execute = "auto"  # auto, always, never
nbsphinx_execute_arguments = [
    "--InlineBackend.rc=figure.facecolor='w'",
    "--InlineBackend.rc=font.size=15",
]

# sphinxcontrib-bibtex configuration
bibtex_bibfiles = ["bibliography.bib"]


def linkcode_resolve(domain, info):
    """Determine the URL corresponding to Python object.

    This is taken from SciPy's conf.py:
    https://github.com/scipy/scipy/blob/master/doc/source/conf.py.
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        lineno = None

    if lineno:
        linespec = "#L%d-L%d" % (lineno, lineno + len(source) - 1)
    else:
        linespec = ""

    startdir = os.path.abspath(os.path.join(dirname(__file__), ".."))
    fn = relpath(fn, start=startdir).replace(os.path.sep, "/")

    if fn.startswith("orix/"):
        m = re.match(r"^.*dev0\+([a-f0-9]+)$", __version__)
        pre_link = "https://github.com/pyxem/orix/blob/"
        if m:
            return pre_link + "%s/%s%s" % (m.group(1), fn, linespec)
        elif "dev" in __version__:
            return pre_link + "master/%s%s" % (fn, linespec)
        else:
            return pre_link + "v%s/%s%s" % (__version__, fn, linespec)
    else:
        return None



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from orix._util import deprecated


class TestDeprecationWarning:
    def test_deprecation_since(self):
        """Ensure functions decorated with the custom deprecated
        decorator returns desired output, raises a desired warning, and
        gets the desired additions to their docstring.
        """

        class Foo:
            @property
            @deprecated(
                since="1.3",
                alternative="bar_prop2",
                removal="1.4",
                object_type="not a function",
            )
            def bar_prop(self):
                return 1

            @deprecated(since=0.7, alternative="bar_func3", removal=0.8)
            def bar_func1(self, n):
                """Some docstring."""
                return n + 1

            @deprecated(since=1.9)
            def bar_func2(self, n):
                """Another docstring.

                Notes
                -----
                Some existing notes.
                """
                return n + 2

        my_foo = Foo()

        with pytest.warns(np.VisibleDeprecationWarning) as record:
            assert my_foo.bar_func1(4) == 5
        desired_msg = (
            "Function `bar_func1()` is deprecated and will be removed in version 0.8. "
            "Use `bar_func3()` instead."
        )
        assert str(record[0].message) == desired_msg
        assert my_foo.bar_func1.__doc__ == (
            "[*Deprecated*] Some docstring.\n"
            "\nNotes\n-----\n"
            ".. deprecated:: 0.7\n"
            f"   {desired_msg}"
        )

        with pytest.warns(np.VisibleDeprecationWarning) as record2:
            assert my_foo.bar_func2(4) == 6
        desired_msg2 = "Function `bar_func2()` is deprecated."
        assert str(record2[0].message) == desired_msg2
        assert my_foo.bar_func2.__doc__ == (
            "[*Deprecated*] Another docstring.\n"
            "\nNotes\n-----\n"
            "Some existing notes.\n\n"
            ".. deprecated:: 1.9\n"
            f"   {desired_msg2}"
        )

        with pytest.warns(np.VisibleDeprecationWarning) as record3:
            assert my_foo.bar_prop == 1
        desired_msg3 = (
            "Property `bar_prop` is deprecated and will be removed in version 1.4. "
            "Use `bar_prop2` instead."
        )
        assert str(record3[0].message) == desired_msg3
        assert my_foo.__class__.bar_prop.__doc__ == (
            "[*Deprecated*] \n"
            "\nNotes\n-----\n"
            ".. deprecated:: 1.3\n"
            f"   {desired_msg3}"
        )



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from packaging import version

from diffpy.structure import Lattice, Structure
from matplotlib import __version__ as _MPL_VERSION
from matplotlib import pyplot as plt
import numpy as np
import pytest

from orix.plot._util import Arrow3D
from orix.plot.unit_cell_plot import (
    _calculate_basic_unit_cell_edges,
    _calculate_basic_unit_cell_vertices,
)
from orix.quaternion import Orientation


def test_unit_cell_plot_default():
    ori = Orientation.random()
    fig = ori.plot_unit_cell(return_figure=True)
    assert len(fig.axes) == 1
    axes = fig.axes[0]
    assert len(axes.lines) == 12  # 12 edges in orthorhombic unit cell
    # 6 Arrow3D -> 3 for both sample and crystal reference frames
    if version.parse(_MPL_VERSION) >= version.parse("3.4"):  # pragma: no cover
        assert len(axes.patches) == 6
    else:  # pragma: no cover
        assert len(axes.artists) == 6
    # test default projection
    assert axes.azim == -90
    assert round(axes.elev) == 90
    plt.close("all")

    plt.close("all")


def test_unit_cell_plot_multiple_orientations_raises():
    ori = Orientation.random((2,))
    with pytest.raises(ValueError, match="Can only plot a single unit cell"):
        ori.plot_unit_cell()
    plt.close("all")


def test_unit_cell_plot_orthorhombic():
    ori = Orientation.random()
    lattice = Lattice(1, 2, 3, 90, 90, 90)
    structure = Structure(lattice=lattice)
    _ = ori.plot_unit_cell(return_figure=True, structure=structure)


def test_unit_cell_plot_hexagonal():
    ori = Orientation.random()
    lattice = Lattice(1, 1, 2, 90, 90, 120)
    structure = Structure(lattice=lattice)
    fig = ori.plot_unit_cell(return_figure=True, structure=structure)
    axes = fig.axes[0]
    # should only be 12 edges in hexagonal unit cell, this test checks
    # that the edges parallel to (0000)-(11-20) are not plotted
    assert len(axes.lines) == 12
    plt.close("all")


def test_unit_cell_plot_crystal_reference_axes_position_center():
    ori = Orientation.identity()
    a1, a2, a3 = 1, 1.5, 2
    lattice = Lattice(a1, a2, a3, 90, 90, 90)
    structure = Structure(lattice=lattice)
    # test cell center
    fig = ori.plot_unit_cell(
        return_figure=True,
        structure=structure,
        crystal_axes_loc="center",
    )
    if version.parse(_MPL_VERSION) >= version.parse("3.4"):  # pragma: no cover
        arrows = fig.axes[0].patches
    else:  # pragma: no cover
        arrows = fig.axes[0].artists
    crys_ref_ax = [p for p in arrows if "Crystal reference axes" in p.get_label()]
    crys_ref_ax_data = np.stack([np.array(a._verts3d) for a in crys_ref_ax])
    assert np.allclose(crys_ref_ax_data[:, :, 0], 0)
    plt.close("all")


def test_unit_cell_plot_crystal_reference_axes_position_origin():
    ori = Orientation.identity()
    a1, a2, a3 = 1, 1.5, 2
    lattice = Lattice(a1, a2, a3, 90, 90, 90)
    structure = Structure(lattice=lattice)
    # test cell center
    fig = ori.plot_unit_cell(
        return_figure=True,
        structure=structure,
        crystal_axes_loc="origin",
    )
    if version.parse(_MPL_VERSION) >= version.parse("3.4"):  # pragma: no cover
        arrows = fig.axes[0].patches
    else:  # pragma: no cover
        arrows = fig.axes[0].artists
    crys_ref_ax = [p for p in arrows if "Crystal reference axes" in p.get_label()]
    crys_ref_ax_data = np.stack([np.array(a._verts3d) for a in crys_ref_ax])
    assert np.allclose(crys_ref_ax_data[:, :, 0] + np.array((a1, a2, a3)) / 2, 0)
    plt.close("all")


def test_unit_cell_plot_crystal_reference_axes_position_raises():
    ori = Orientation.identity()
    with pytest.raises(ValueError, match="Crystal_axes_loc must be either"):
        ori.plot_unit_cell(crystal_axes_loc="test")
    plt.close("all")


def test_calculate_basic_unit_cell_raises():
    lattice = Lattice()
    with pytest.raises(ValueError, match=r"Vectors must be \(3, 3\) array."):
        _ = _calculate_basic_unit_cell_edges(lattice.base, np.ones((3, 4)))

    with pytest.raises(ValueError, match=r"Vectors must be \(3, 3\) array."):
        _ = _calculate_basic_unit_cell_vertices(np.ones((3, 4)))


def test_unit_cell_plot_invalid_structure_raises():
    ori = Orientation.random((1,))
    with pytest.raises(TypeError, match=r"Structure must be diffpy.structure."):
        ori.plot_unit_cell(structure=np.arange(3))


def test_arrow3D():
    _, ax = plt.subplots(subplot_kw=dict(projection="3d"))
    a = Arrow3D((0, 1), (0, 1), (0, 1), arrowstyle="-|>", mutation_scale=20)
    ax.add_artist(a)
    plt.draw()
    plt.close("all")



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from orix.quaternion.symmetry import Ci, Oh
from orix.vector import FundamentalSector, Vector3d


class TestFundamentalSector:
    # Most of the FundamentalSector class is tested in test_symmetry.py
    def test_center_from_s2_sampling(self):
        v = Vector3d([[0.5, 0, 0], [0, 0.5, 0.5], [0, 0, 1], [1, 1, 0], [0, 1, 1]])
        fs = FundamentalSector(v)

        assert fs.vertices.size == 6
        assert np.allclose(fs.center.data, [[0.534, 0.322, 0.534]], atol=1e-3)

    def test_edges(self):
        edge_steps = 1000
        fs1 = Ci.fundamental_sector
        assert np.allclose(
            fs1.edges.data, Vector3d.zvector().get_circle(steps=edge_steps).data
        )

        # Make sure that desired parts of the fundamental sector edge
        # are part of the actual edge
        fs2 = Oh.fundamental_sector
        circles = fs2.get_circle(steps=edge_steps)
        desired_edges = circles[circles <= fs2]
        actual_edges = fs2.edges
        actual_edges_data = actual_edges.data
        for edge in desired_edges.data:
            assert np.all(np.isclose(edge, actual_edges_data), axis=1).any()

        # "Joints" between the three edges are included as well
        assert desired_edges.size == 375
        assert actual_edges.size == 378



# -*- coding: utf-8 -*-
# Copyright 2018-2022 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions and classes for managing orix.

This module and documentation is only relevant for orix developers, not
for users.

.. warning:
    This module and its submodules are for internal use only.  Do not
    use them in your own code. We may change the API at any time with no
    warning.
"""


import functools
import inspect
import warnings

import numpy as np


class deprecated:
    """Decorator to mark deprecated functions or properties with an
    informative warning.

    Adapted from
    `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py#L297>`_
    and `matplotlib
    <https://github.com/matplotlib/matplotlib/blob/master/lib/matplotlib/_api/deprecation.py#L122>`_.
    """

    def __init__(self, since, alternative=None, removal=None, object_type="function"):
        """Visible deprecation warning.

        Parameters
        ----------
        since : str, int or float
            The release at which this API became deprecated.
        alternative : str, optional
            An alternative API that the user may use in place of the
            deprecated API.
        removal : str, int or float, optional
            The expected removal version.
        object_type : str, optional
            Type of the deprecated object, either "function" or
            "property".
        """
        self.since = since
        self.alternative = alternative
        self.removal = removal
        self.object_type = object_type

    def __call__(self, func):
        # Wrap function or property to raise warning when called, and
        # add warning to docstring
        if self.object_type.lower() != "function":
            object_type = "Property"
            parentheses = ""
        else:
            object_type = "Function"
            parentheses = "()"
        if self.alternative is not None:
            alt_msg = f" Use `{self.alternative}{parentheses}` instead."
        else:
            alt_msg = ""
        if self.removal is not None:
            rm_msg = f" and will be removed in version {self.removal}"
        else:
            rm_msg = ""
        msg = (
            f"{object_type} `{func.__name__}{parentheses}` is deprecated"
            f"{rm_msg}.{alt_msg}"
        )

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.simplefilter(
                action="always", category=np.VisibleDeprecationWarning
            )
            func_code = func.__code__
            warnings.warn_explicit(
                message=msg,
                category=np.VisibleDeprecationWarning,
                filename=func_code.co_filename,
                lineno=func_code.co_firstlineno + 1,
            )
            return func(*args, **kwargs)

        # Modify docstring to display deprecation warning
        old_doc = inspect.cleandoc(func.__doc__ or "").strip("\n")
        notes_header = "\nNotes\n-----"
        new_doc = (
            f"[*Deprecated*] {old_doc}\n"
            f"{notes_header if notes_header not in old_doc else ''}\n"
            f".. deprecated:: {self.since}\n"
            f"   {msg.strip()}"  # Matplotlib uses three spaces
        )
        wrapped.__doc__ = new_doc

        return wrapped
